
using namespace std;
#include <chrono>
#include <algorithm>
//#include <string>
void print_policy(int t, int i_policy, problem_state* mp);
void backward_updating(problem_state* mp, double** per_period_updating_times, int improvement_round, bool record_times);

class mycallback_dynamic : public GRBCallback
{
public:
	int numvars;
	GRBVar* vars;
	ofstream* logfile;
	problem_state* xmp;
	bool* xwhich_TD;
	int* xactive_nodes_indices;
	int xnb_active_nodes;
	std::chrono::steady_clock::time_point xstart_moment;

	mycallback_dynamic(int xnumvars, GRBVar* xvars, ofstream* xlogfile,
		problem_state* mp, bool* which_TD, int* active_nodes_indices,
		int nb_active_nodes, std::chrono::steady_clock::time_point start_moment) {
		numvars = xnumvars;
		vars = xvars;
		logfile = xlogfile;
		xmp = mp;
		xwhich_TD = which_TD;
		xactive_nodes_indices = active_nodes_indices;
		xnb_active_nodes = nb_active_nodes;
		xstart_moment = start_moment;
	}
protected:
	void callback() {
		try {
			if (where == GRB_CB_POLLING) {
			}
			else if (where == GRB_CB_PRESOLVE) {
			}
			else if (where == GRB_CB_SIMPLEX) {
			}
			else if (where == GRB_CB_MIP) {
				// MIP solution callback
			}
			else if (where == GRB_CB_MIPSOL) {
				// MIP solution callback
				cout << "MIP new solution callback";
				int nodecnt = (int)getDoubleInfo(GRB_CB_MIPSOL_NODCNT);
				double obj = getDoubleInfo(GRB_CB_MIPSOL_OBJ);
				double obj_bound = getDoubleInfo(GRB_CB_MIPSOL_OBJBND);
				int solcnt = getIntInfo(GRB_CB_MIPSOL_SOLCNT);
				double* x = getSolution(vars, numvars);

				// Record how many decisions were there before
				int* old_number_of_policies = new int[xmp->T];
				for (int t = 0; t < xmp->T; t++) {
					old_number_of_policies[t] = xmp->how_many_dec_there[t];
				}

				// Export of decisions
				int offset_active_node;

				for (int i_node = 1; i_node < xmp->nb_nodes - 1; i_node++) {
					if (xwhich_TD[i_node]) {
						int t = xmp->nodes[i_node].time;
						offset_active_node = xactive_nodes_indices[i_node];

						// Check if maybe such a decision is there already
						bool unique_or_not = true;
						int variable_index = 0;

						for (int i_policy_to_compare = 0; i_policy_to_compare < xmp->how_many_dec_there[t]; i_policy_to_compare++) {
							int the_same = 0;
							for (int i_f = 0; i_f < xmp->nb_locations; i_f++) {
								variable_index = i_f * xnb_active_nodes + offset_active_node;
								if (abs(x[variable_index] - xmp->policies_x[t][i_policy_to_compare][i_f]) < 0.1) {
									the_same++;
								}
							}

							if (the_same > xmp->nb_locations - 1) {
								unique_or_not = false;
								break;
							}
						}

						if (unique_or_not) {
							for (int i_f = 0; i_f < xmp->nb_locations; i_f++) {
								variable_index = i_f * xnb_active_nodes + offset_active_node;
								//std::cout << t << " " << xmp->how_many_dec_there[t] << " " << xmp->policies_x[t][xmp->how_many_dec_there[t]][i_f] << " " << variable_index;// << abs(round(x[variable_index])) << "\n";
								xmp->policies_x[t][xmp->how_many_dec_there[t]][i_f] = abs(round(x[variable_index]));
							}
							xmp->how_many_dec_there[t]++;
						}
					}
				}

				// Run the backward recursion
				double** xper_period_updating_times;
				xper_period_updating_times = new double* [1];
				backward_updating(xmp, xper_period_updating_times, 0, false);
				std::chrono::steady_clock::time_point xend_moment;
				xend_moment = std::chrono::steady_clock::now();
				double time_till_now;
				time_till_now = (double)(std::chrono::duration_cast<std::chrono::milliseconds>(xend_moment - xstart_moment).count()) / 1000;

				// Write the outcome to the file
				*logfile << time_till_now << " " << xmp->TBD_nodes[0].han_plus_cost_to_go << " " << obj_bound << "\n";

				std::cout << "Numbers of policies before: ";
				for (int t = 0; t < xmp->T; t++) {
					std::cout << xmp->how_many_dec_there[t] << " ";
				}
				std::cout << "\n";

				// Clean the "here and now decision computed" marker for TBD nodes
				for (int t = 0; t < xmp->T; t++) {
					for (int i_dr = 0; i_dr < t + 2; i_dr++) {
						for (int i_dec = old_number_of_policies[t]; i_dec < xmp->how_many_dec_there[t]; i_dec++) {
							xmp->TBD_nodes[xmp->map_TBD_to_node[t][i_dr][i_dec]].han_cost_computed = false;
							xmp->TBD_nodes[xmp->map_TBD_to_node[t][i_dr][i_dec]].han_cost_computed_no_IU = false;
						}
					}
				}

				// Clean the list of policies
				std::cout << "Numbers of policies after: ";
				for (int t = 0; t < xmp->T; t++) {
					xmp->how_many_dec_there[t] = old_number_of_policies[t];
					std::cout << old_number_of_policies[t] << " ";
				}
				std::cout << "\n";

				delete[] x;
			}
			else if (where == GRB_CB_MIPNODE) {
			}
			else if (where == GRB_CB_BARRIER) {
			}
			else if (where == GRB_CB_MESSAGE) {
			}
		}
		catch (GRBException e) {
			cout << "Error number: " << e.getErrorCode() << endl;
			cout << e.getMessage() << endl;
		}
		catch (...) {
			cout << "Error during callback" << endl;
		}
	}
};

double node_HAN_costs_CCG(int TBD_node, problem_state* mp){
	//std::cout << "============= Here-and-now cost evaluation of a node =========\n";
	int t = mp->TBD_nodes[TBD_node].time;
	int i_decision = mp->TBD_nodes[TBD_node].dec_number;
	double to_return = 0;

	try {
		GRBModel model = GRBModel (*mp->env);
		model.set(GRB_StringAttr_ModelName, "Facility planning");
		model.set(GRB_IntAttr_ModelSense, GRB_MINIMIZE);
		model.set(GRB_DoubleParam_TimeLimit, mp->TimeLimit);
		model.set(GRB_DoubleParam_MIPGap, 0.001);
		model.set(GRB_IntParam_OutputFlag, 0);
		model.set(GRB_IntParam_Threads, 4);

		// Variable x which facilities are open
		GRBVar* x;
		x = model.addVars(mp->nb_locations, GRB_BINARY);

		// Initialize the number of vertices of the uncertainty set
		const int nb_vertices = 6;
		double vertices[6][4] = {
			{1 - mp->uncertainty_level, 1 - mp->uncertainty_level, 1 + mp->uncertainty_level, 1 + mp->uncertainty_level},
			{1 - mp->uncertainty_level, 1 + mp->uncertainty_level, 1 - mp->uncertainty_level, 1 + mp->uncertainty_level},
			{1 + mp->uncertainty_level, 1 - mp->uncertainty_level, 1 - mp->uncertainty_level, 1 + mp->uncertainty_level},
			{1 - mp->uncertainty_level, 1 + mp->uncertainty_level, 1 + mp->uncertainty_level, 1 - mp->uncertainty_level},
			{1 + mp->uncertainty_level, 1 - mp->uncertainty_level, 1 + mp->uncertainty_level, 1 - mp->uncertainty_level},
			{1 + mp->uncertainty_level, 1 + mp->uncertainty_level, 1 - mp->uncertainty_level, 1 - mp->uncertainty_level},
		};

		// Variable y_0 - fixed delivery values
		GRBVar*** y_0;
		y_0 = new GRBVar **[mp->nb_customers];
		for (int i_c = 0; i_c < mp->nb_customers; i_c++) {
			y_0[i_c] = new GRBVar *[mp->nb_locations];

			for (int i_l = 0; i_l < mp->nb_locations; i_l++) {
				y_0[i_c][i_l] = model.addVars(nb_vertices, GRB_CONTINUOUS);
			}
		}

		// Variable w_0 - stands for unsatisfied demand fixed
		GRBVar** w_0;
		w_0 = new GRBVar *[mp->nb_customers]; //model.addVars(mp->nb_customers, GRB_CONTINUOUS);

		for (int i_c = 0; i_c < mp->nb_customers; i_c++) {
			w_0[i_c] = model.addVars(nb_vertices, GRB_CONTINUOUS);
		}

		GRBVar ub_cost = model.addVar(-GRB_INFINITY,  GRB_INFINITY, 0.0,  GRB_CONTINUOUS , "ub_cost");
		model.setObjective(ub_cost - 0.0);
		GRBLinExpr expr_constant = 0;

		for (int i_l = 0; i_l < mp->nb_locations; i_l++) {// Fixing the value of the location variables
			model.addConstr(x[i_l] == mp->policies_x[mp->TBD_nodes[TBD_node].time][i_decision][i_l]);
		}

		// OK so now starts the real thing
		int offset_RC = 0;

		// Determine the time period and demand regime of the node
		double p = mp->TBD_nodes[TBD_node].total_demand;
		for (int i_c = 0; i_c < mp->nb_factors; i_c++) {
			mp->demand_factors[i_c].ub_demand = (1 + mp->uncertainty_level) * (p / mp->nb_factors);
			mp->demand_factors[i_c].lb_demand = (1 - mp->uncertainty_level) * (p / mp->nb_factors);
		}
		
		double demand_in_vertex_customer;

		// Capacity constraints
		for (int i_vertex = 0; i_vertex < nb_vertices; i_vertex++){
			for (int i_l = 0; i_l < mp->nb_locations; i_l++) {// Zero the linear expressions
				expr_constant = - x[i_l] * mp->factories[i_l].capacity;

				// Build up constraint material
				for (int i_c = 0; i_c < mp->nb_customers; i_c++) {
					expr_constant += y_0[i_c][i_l][i_vertex];
				}

				model.addConstr(expr_constant <= 0.0);
			}
		}

		// w == demand minus delivery
		for (int i_vertex = 0; i_vertex < nb_vertices; i_vertex++){
			for (int i_c = 0; i_c < mp->nb_customers; i_c++) {// Zero the linear expressions
				demand_in_vertex_customer = 0;

				for (int i_f = 0; i_f < mp->nb_factors; i_f++) {
					demand_in_vertex_customer += mp->Phi[i_c][i_f] * (p / mp->nb_factors) * vertices[i_vertex][i_f];
				}

				expr_constant = - w_0[i_c][i_vertex] + demand_in_vertex_customer;

				for (int i_l = 0; i_l < mp->nb_locations; i_l++) {
					expr_constant -= y_0[i_c][i_l][i_vertex];
				}

				model.addConstr(expr_constant == 0.0);
			}
		}

		// Now the constraint on the total cost in a given node
		for (int i_vertex = 0; i_vertex < nb_vertices; i_vertex++){
			expr_constant = - ub_cost;

			for (int i_l = 0; i_l < mp->nb_locations; i_l++) {
				expr_constant += x[i_l] * mp->m_cost;
			}

			for (int i_c = 0; i_c < mp->nb_customers; i_c++) {
				expr_constant += w_0[i_c][i_vertex] * mp->b;

				for (int i_l = 0; i_l < mp->nb_locations; i_l++) {
					expr_constant += y_0[i_c][i_l][i_vertex] * (mp->c[i_l][i_c] - mp->revenue);
				}
			}

			model.addConstr(expr_constant <= 0.0);
		}

		model.optimize();
		to_return = ub_cost.get(GRB_DoubleAttr_X);
		/////////////////////////////////////////////////////////////////////////////////
		// Condition if the total costs in this vertex equal to the node totals
		for (int i_vertex = 0; i_vertex < nb_vertices; i_vertex++){
			double max_minus_node = - ub_cost[i_node].get(GRB_DoubleAttr_X);

			for (int i_l = 0; i_l < mp->nb_locations; i_l++) {
				max_minus_node += x[i_l].get(GRB_DoubleAttr_X) * mp->m_cost;
			}

			for (int i_c = 0; i_c < mp->nb_customers; i_c++) {
				max_minus_node += w_0[i_c][i_vertex].get(GRB_DoubleAttr_X) * mp->b;

				for (int i_l = 0; i_l < mp->nb_locations; i_l++) {
					max_minus_node += y_0[i_c][i_l][i_vertex].get(GRB_DoubleAttr_X) * (mp->c[i_l][i_c] - mp->revenue);
				}
			}

			if (max_minus_node >= -0.01) {// if a given vertex is tight
				//mp->TBD_nodes[TBD_node].which_vertex_worst = i_vertex;
				mp->nodes[mp->map_TB_to_node[mp->TBD_nodes[TBD_node].time][mp->TBD_nodes[TBD_node].i_dr]] = i_vertex;
			}
		}
		//////////////////////////////////////////////////////////////////////////////////

		offset_active_node++;

		delete[] x;
		delete[] y_0;
		delete[] w_0;
	}
	catch (GRBException e) {
		cout << "Error code = " << e.getErrorCode() << endl;
		cout << e.getMessage() << endl;
	}
	catch (...) {
		cout << "Exception during optimization" << endl;
	}

	return to_return;
}

void node_update(int TBD_node, problem_state* mp){
	if (!mp->TBD_nodes[TBD_node].total_cost_computed) {
		int TB_node; // to indentify the index of the corresponding TB node
		if (mp->TBD_nodes[TBD_node].time > -1) {
			TB_node = mp->map_TB_to_node[mp->TBD_nodes[TBD_node].time][mp->TBD_nodes[TBD_node].i_dr];
		}
		else {
			TB_node = 0;
		}

		int nb_of_children_i_dr = mp->nodes[TB_node].nb_of_children;
		int time = mp->TBD_nodes[TBD_node].time;
		int i_dr = mp->TBD_nodes[TBD_node].i_dr;

		// Compute here-and-now cost
		//if (!mp->TBD_nodes[TBD_node].han_cost_computed){ // Only if the here and now not computed yet
		if((time > -1) && !(mp->TBD_nodes[TBD_node].han_cost_computed)){
			mp->TBD_nodes[TBD_node].han_cost = node_HAN_costs_CCG(TBD_node, mp);
			mp->TBD_nodes[TBD_node].han_cost_computed = true;
		}

		// Now the dilemma on the here-and-now + cost to go
		if (mp->TBD_nodes[TBD_node].time == mp->T - 1) { // if last time period
			mp->TBD_nodes[TBD_node].han_plus_cost_to_go = mp->TBD_nodes[TBD_node].han_cost;
		}
		else { // Cost to go estimation
			int nb_of_children_i_dr = mp->nodes[TB_node].nb_of_children; // nb of possible new demand regimes
			int* possible_i_dr = new int[nb_of_children_i_dr]; // possible demand regimes
			double* cost_to_go = new double[nb_of_children_i_dr]; // best decision for each new demand regime
			int* best_next_node = new int[nb_of_children_i_dr]; // For storing the best next node
			double worst_cost_to_go = -pow(10, 6);
			int index_worst_cost_to_go = 0;

			for (int i_TB_child = 0; i_TB_child < mp->nodes[TB_node].nb_of_children; i_TB_child++) {
				int index_TB_child = mp->nodes[TB_node].children[i_TB_child];
				possible_i_dr[i_TB_child] = mp->nodes[index_TB_child].i_dr;
				cost_to_go[i_TB_child] = pow(10, 8); // LARGE NUMBER
			}

			for (int i_child = 0; i_child < mp->TBD_nodes[TBD_node].nb_of_children; i_child++) {
				int TBD_child_node = mp->TBD_nodes[TBD_node].children[i_child];
				// If the decision is actually existing 
				if (mp->TBD_nodes[TBD_child_node].dec_number < mp->how_many_dec_there[mp->TBD_nodes[TBD_child_node].time]){
					bool factories_OK = true;

					// Only for nodes other than the 0 node
					if (TBD_node > 0) {
						for (int i_f = 0; i_f < mp->nb_locations; i_f++) {
							if (mp->policies_x[mp->TBD_nodes[TBD_node].time][mp->TBD_nodes[TBD_node].dec_number][i_f] > mp->policies_x[mp->TBD_nodes[TBD_child_node].time][mp->TBD_nodes[TBD_child_node].dec_number][i_f]){
								factories_OK = false; // if any factor less open later, disregard
							}
						}
					}

					if (factories_OK) {// Identify the right bucket to compare it to
						mp->TBD_nodes[TBD_node].children_possible[i_child] = true;
						int right_idr_index = 0;

						for (int i_index = 0; i_index < nb_of_children_i_dr; i_index++) {
							if (mp->TBD_nodes[TBD_child_node].i_dr == possible_i_dr[i_index]) {
								right_idr_index = i_index;
							}
						}

						if (mp->TBD_nodes[TBD_child_node].han_plus_cost_to_go < cost_to_go[right_idr_index] - pow(10, -1)) {
							cost_to_go[right_idr_index] = mp->TBD_nodes[TBD_child_node].han_plus_cost_to_go;
							best_next_node[right_idr_index] = TBD_child_node;
						}
						else if ((mp->TBD_nodes[TBD_child_node].han_plus_cost_to_go <= cost_to_go[right_idr_index] + pow(10, -1)) && (mp->TBD_nodes[TBD_child_node].han_plus_cost_to_go >= cost_to_go[right_idr_index] - pow(10, -1))) {
							double swap_or_not = (double)rand() / (RAND_MAX);
							if (swap_or_not > 0.1) {
								best_next_node[right_idr_index] = TBD_child_node;
							}
						}
					}
					else {
						mp->TBD_nodes[TBD_node].children_possible[i_child] = false;
					}
				}
				else {
					mp->TBD_nodes[TBD_node].children_possible[i_child] = false;
				}
			}

			// Now choose what is the worst_i_dr
			for (int i_index = 0; i_index < nb_of_children_i_dr; i_index++) {
				if (cost_to_go[i_index] > worst_cost_to_go + pow(10, -1)) {
					worst_cost_to_go = cost_to_go[i_index];
					index_worst_cost_to_go = i_index;
				}
				else if ((cost_to_go[i_index] <= worst_cost_to_go + pow(10, -1)) && (cost_to_go[i_index] >= worst_cost_to_go - pow(10, -1))) {
					double swap_or_not = (double)rand() / (RAND_MAX);
					if (swap_or_not > 0.1) {
						worst_cost_to_go = cost_to_go[i_index];
						index_worst_cost_to_go = i_index;
					}
				}
			}

			// Now update the value
			mp->TBD_nodes[TBD_node].han_plus_cost_to_go = mp->TBD_nodes[TBD_node].han_cost + worst_cost_to_go;
			mp->TBD_nodes[TBD_node].worst_TBD_child = best_next_node[index_worst_cost_to_go];
			//std::cout << "Node[" << TBD_node << "] t: " << mp->TBD_nodes[TBD_node].time << ", i_dr: " << mp->TBD_nodes[TBD_node].i_dr << "\n";
			//std::cout << "Node[" << TBD_node << "] han: " << mp->TBD_nodes[TBD_node].han_cost << ", han plus ctg: " << mp->TBD_nodes[TBD_node].han_plus_cost_to_go << "\n";
			//std::cout << "Child node[" << mp->TBD_nodes[TBD_node].worst_TBD_child << "] t: " << mp->TBD_nodes[mp->TBD_nodes[TBD_node].worst_TBD_child].time << ", i_dr: " << mp->TBD_nodes[mp->TBD_nodes[TBD_node].worst_TBD_child].i_dr << "\n";
			//std::cout << "Child node[" << mp->TBD_nodes[TBD_node].worst_TBD_child << "] han: " << mp->TBD_nodes[mp->TBD_nodes[TBD_node].worst_TBD_child].han_cost << ", han plus ctg: " << mp->TBD_nodes[mp->TBD_nodes[TBD_node].worst_TBD_child].han_plus_cost_to_go << "\n";

			// Now update the list of worst nodes
			if (mp->TBD_nodes[TBD_node].time < mp->T - 1) {
				for (int i_child = 0; i_child < mp->TBD_nodes[TBD_node].nb_of_children; i_child++) {
					int right_idr_index = 0;
					int TBD_child_node = mp->TBD_nodes[TBD_node].children[i_child];

					for (int i_index = 0; i_index < nb_of_children_i_dr; i_index++) {
						if (mp->TBD_nodes[TBD_child_node].i_dr == possible_i_dr[i_index]) {
							right_idr_index = i_index;
						}
					}

					if (mp->TBD_nodes[TBD_node].children_possible[i_child]) {
						if (mp->TBD_nodes[mp->TBD_nodes[TBD_node].children[i_child]].han_plus_cost_to_go > worst_cost_to_go - pow(10, -1)) {
							//if (cost_to_go[right_idr_index] > worst_cost_to_go - pow(10, -1)) {
								mp->TBD_nodes[TBD_node].children_worst[i_child] = true;
							//}
						}
					}
				}
			}

			delete[] possible_i_dr;
			delete[] best_next_node;
			delete[] cost_to_go;
		}
	}
}

void print_TBD_node(int TBD_node, problem_state* mp, bool to_file){
	if (to_file) {
		std::ofstream my_file;
		my_file.open(mp->output_file.c_str(), std::ios::app);
		//my_file.open(mp->output_file.c_str());
		my_file << "TBD_node: " << TBD_node << " T= " << mp->TBD_nodes[TBD_node].time << " B= " << mp->TBD_nodes[TBD_node].i_dr << " i_dec= " << mp->TBD_nodes[TBD_node].dec_number << "\n";
		my_file << "Here  cost: " << mp->TBD_nodes[TBD_node].han_cost << "\n";
		my_file << "Total cost: " << mp->TBD_nodes[TBD_node].han_plus_cost_to_go << "\n";
		my_file << "Here  cost no IU: " << mp->TBD_nodes[TBD_node].han_cost_no_IU << "\n";
		my_file << "Total cost no IU: " << mp->TBD_nodes[TBD_node].han_plus_cost_to_go_no_IU << "\n";

		if (mp->TBD_nodes[TBD_node].time < mp->T - 1)
		{
			my_file << "Worst next idr : " << mp->TBD_nodes[mp->TBD_nodes[TBD_node].worst_TBD_child].i_dr << "\n";
			my_file << "Worst next node: " << mp->TBD_nodes[TBD_node].worst_TBD_child << "\n";
		}
		my_file.close();
	}
	else {
		//std::cout << "TBD_node: " << TBD_node << " T= " << mp->TBD_nodes[TBD_node].time << " B= " << mp->TBD_nodes[TBD_node].i_dr << " i_dec= " << mp->TBD_nodes[TBD_node].dec_number << "\n";
		//std::cout << "Here  cost: " << mp->TBD_nodes[TBD_node].han_cost << "\n";
		//std::cout << "Total cost: " << mp->TBD_nodes[TBD_node].han_plus_cost_to_go << "\n";
		//std::cout << "Here  cost no IU: " << mp->TBD_nodes[TBD_node].han_cost_no_IU << "\n";
		//std::cout << "Total cost no IU: " << mp->TBD_nodes[TBD_node].han_plus_cost_to_go_no_IU << "\n";

		if (mp->TBD_nodes[TBD_node].time < mp->T - 1)
		{
			//std::cout << "Worst next idr : " << mp->TBD_nodes[mp->TBD_nodes[TBD_node].worst_TBD_child].i_dr << "\n";
			//std::cout << "Worst next node: " << mp->TBD_nodes[TBD_node].worst_TBD_child << "\n";
		}
	}
}

void print_all_TBD_nodes(problem_state* mp, bool to_file){
	print_TBD_node(0, mp, to_file);

	for (int t = 0; t < mp->T; t++){
		for (int i_dr = 0; i_dr < mp->how_many_i_dr[t]; i_dr++){
			for (int i_dec = 0; i_dec < mp->how_many_dec_there[t]; i_dec++){
				int TBD_node = mp->map_TBD_to_node[t][i_dr][i_dec];
				print_TBD_node(TBD_node, mp, to_file);
			}
		}
	}
}

void backward_updating(problem_state* mp, double** per_period_updating_times, int improvement_round, bool record_times){
	std::cout << "backward updating\n";
	std::chrono::steady_clock::time_point start_time;// = std::chrono::steady_clock::now();
	std::chrono::steady_clock::time_point end_time;// = std::chrono::steady_clock::now();

	// updating the nodes going backwards
	for (int t = mp->T - 1; t >= 0; t--){
		start_time = std::chrono::steady_clock::now();
		for (int i_dr = 0; i_dr < mp->how_many_i_dr[t]; i_dr++){
			for (int i_dec = 0; i_dec < mp->how_many_dec_there[t]; i_dec++){
				int TBD_node = mp->map_TBD_to_node[t][i_dr][i_dec];
				node_update(TBD_node, mp);
			}
		}
		end_time = std::chrono::steady_clock::now();

		double solution_time = (double)(std::chrono::duration_cast<std::chrono::milliseconds>(end_time - start_time).count()) / 1000;
		if (record_times) {
			per_period_updating_times[improvement_round][t] = solution_time;
		}
	}

	node_update(0, mp);
}

void print_policy(int t, int i_policy, problem_state* mp, bool to_file) {
	if (to_file) {
		std::ofstream my_file;
		my_file.open(mp->output_file.c_str(), std::ios::app);
		my_file << "Printing policy t=" << t << " i_policy= " << i_policy << "\n";

		for (int i_l = 0; i_l < mp->nb_locations; i_l++) {
			my_file << mp->policies_x[t][i_policy][i_l] << " ";
		}
		my_file << "\n";
		my_file.close();
	}
	else {
		for (int i_l = 0; i_l < mp->nb_locations; i_l++) {
			std::cout << mp->policies_x[t][i_policy][i_l] << " ";
		}
		std::cout << "\n";
	}
}

void print_all_policies(problem_state* mp, bool to_file) {
	std::cout << "Printing all policies\n";

	for (int t = 0; t < mp->T; t++) {
		for (int i_policy = 0; i_policy < mp->how_many_dec_there[t]; i_policy++) {
			print_policy(t, i_policy, mp, to_file);
		}
		std::cout << "\n";
	}

	std::cout << "\n";
}

void evaluate_all_trajectories_and_sort(problem_state* mp, int** trajectories_list, double *trajectories_list_values, int*trajectories_list_indices, bool*which_TD, double how_many_first, bool IU_included, int worst_TD_method, int*how_many_were_used) {
	for (int i_node = 1; i_node < mp->nb_nodes - 1; i_node++) {// First clean the logical array
		which_TD[i_node] = false;
	}
	
	for (int i_trajectory = 0; i_trajectory < pow(2, mp->T); i_trajectory++) {
		int time = -1;
		int i_dr = 0;
		int TBD_node = 0;
		double costs_so_far = 0;

		while (time < mp->T - 1) {
			time++;
			i_dr += trajectories_list[i_trajectory][time];
			int best_child_index = 0;
			double best_child_value = pow(10, 8);

			for (int i_child = 0; i_child < mp->TBD_nodes[TBD_node].nb_of_children; i_child++) {
				int TBD_child_node = mp->TBD_nodes[TBD_node].children[i_child];

				if (mp->TBD_nodes[TBD_node].children_possible[i_child] && (mp->TBD_nodes[TBD_child_node].i_dr == i_dr)) {
					if(IU_included){
						if (mp->TBD_nodes[TBD_child_node].han_plus_cost_to_go < best_child_value) {
							best_child_index = TBD_child_node;
							best_child_value = mp->TBD_nodes[TBD_child_node].han_plus_cost_to_go;
						}
					} else {
						if (mp->TBD_nodes[TBD_child_node].han_plus_cost_to_go_no_IU < best_child_value) {
							best_child_index = TBD_child_node;
							best_child_value = mp->TBD_nodes[TBD_child_node].han_plus_cost_to_go_no_IU;
						}
					}
				}
			}

			if(IU_included){
				costs_so_far += mp->TBD_nodes[best_child_index].han_cost;
			} else {
				costs_so_far += mp->TBD_nodes[best_child_index].han_cost_no_IU;
			}
			
			TBD_node = best_child_index;
		}

		trajectories_list_values[i_trajectory] = costs_so_far;
	}

	// Now the sorting comes
	int i_buffer = 0;
	double d_buffer = 0;

	for (int i_trajectory = 0; i_trajectory < pow(2, mp->T) - 1; i_trajectory++) {
		for (int i_t_2 = pow(2, mp->T) - 1; i_t_2 > i_trajectory; i_t_2--) {
			if (trajectories_list_values[i_t_2] > trajectories_list_values[i_trajectory]) { // Swapping operations
				i_buffer = trajectories_list_indices[i_t_2];
				d_buffer = trajectories_list_values[i_t_2];

				trajectories_list_indices[i_t_2] = trajectories_list_indices[i_trajectory];
				trajectories_list_values[i_t_2] = trajectories_list_values[i_trajectory];

				trajectories_list_indices[i_trajectory] = i_buffer;
				trajectories_list_values[i_trajectory] = d_buffer;
			}
		}
	}

	// Counting how many different trajectories are worst-possible
	for (int i_trajectory = 0; i_trajectory < pow(2, mp->T); i_trajectory++) {
		std::cout << "Trajectory: " << trajectories_list_indices[i_trajectory] << ", value: " << trajectories_list_values[i_trajectory] << "\n";
	}

	double most_expensive_trajectory = trajectories_list_values[0];
	int nb_worst_trajectories = 0;

	// Counting how many different trajectories are worst-possible
	for (int i_trajectory = 0; i_trajectory < pow(2, mp->T); i_trajectory++) {
		if (trajectories_list_values[i_trajectory] > most_expensive_trajectory - 10) {
			nb_worst_trajectories++;
		} else {
			break;
		}
	}

	// Now marking the percentile in the which_TD map
	int how_many_will_be_taken = 0;

	switch (worst_TD_method) {
		case 0:
			how_many_will_be_taken = 1;
			break;
		case 1:
			if (nb_worst_trajectories < how_many_first) {
				how_many_will_be_taken = nb_worst_trajectories;
			} else {
				how_many_will_be_taken = how_many_first;
			}
			break;
		case 2:
			how_many_will_be_taken = nb_worst_trajectories;
			break;
		case 3:
			if (nb_worst_trajectories > how_many_first) {
				how_many_will_be_taken = nb_worst_trajectories;
			}
			else {
				how_many_will_be_taken = how_many_first;
			}
			break;
		case 4:
			how_many_will_be_taken = how_many_first;
			break;
	}

	*how_many_were_used = how_many_will_be_taken;

	// This is my working area now
	for (int i_trajectory = 0; i_trajectory < how_many_will_be_taken; i_trajectory++) {
		int trajectory_index = trajectories_list_indices[i_trajectory];
		int i_dr = 0;

		for (int t = 0; t < mp->T; t++) {
			std::cout << trajectories_list[trajectory_index][t] << " ";
			i_dr += trajectories_list[trajectory_index][t];
			which_TD[mp->map_TB_to_node[t][i_dr]] = true;
		}

		std::cout << "\n";
	}

	//Printing the results
	for (int t = 0; t < mp->T; t++) {
		for (int i_dr = 0; i_dr < mp->how_many_i_dr[t]; i_dr++) {
			std::cout << which_TD[mp->map_TB_to_node[t][i_dr]] << " ";
		}
		std::cout << "\n";
	}
}

void zero_a_problem(problem_state* mp, bool* which_TD, double* start_facilities, double* end_facilities) {

	for (int i_node = 1; i_node < mp->nb_nodes - 1; i_node++) {
		which_TD[i_node] = true;
	}

	for (int t = 0; t < mp->T; t++) {
		for (int i_policies = 0; i_policies < mp->how_many_dec_there[t]; i_policies++) {
			for (int i_f = 0; i_f < mp->nb_locations; i_f++) {
				mp->policies_x[t][i_policies][i_f] = 0;
			}
		}
	}

	for (int t = 0; t < mp->T; t++) {
		mp->how_many_dec_there[t] = 0;
	}

	mp->TBD_nodes[0].total_cost_computed = false;
	mp->TBD_nodes[0].han_cost = 0;
	mp->TBD_nodes[0].han_plus_cost_to_go = 0;
	mp->TBD_nodes[0].worst_TBD_child = 0;

	for (int i_TBD_node = 1; i_TBD_node < mp->nb_TBD_nodes - 1; i_TBD_node++) {
		mp->TBD_nodes[i_TBD_node].han_cost_computed = false;
		mp->TBD_nodes[i_TBD_node].total_cost_computed = false;
		mp->TBD_nodes[i_TBD_node].han_cost = 0;
		mp->TBD_nodes[i_TBD_node].han_plus_cost_to_go = 0;
		mp->TBD_nodes[i_TBD_node].worst_TBD_child = 0;
	}

	for (int i_f = 0; i_f < mp->nb_locations; i_f++) {
		start_facilities[i_f] = 0;
		end_facilities[i_f] = 0;
	}

}

double dynamic_solution_no_IU(bool* which_TD, int start_period, int end_period, double* start_facilities, double* end_facilities, problem_state* mp, int output_flag, bool AddPolicies) {
	std::cout << "============= Solving the dynamic problem no IU =========\n";
	int nb_active_nodes = 0;
	int* active_nodes_indices = new int[mp->nb_nodes];
	//std::cout << "Backlogging: " << mp->b << " maintenance: " << mp->m_cost << " revenue: " << mp->revenue << "\n";

	for (int i_node = 1; i_node < mp->nb_nodes - 1; i_node++) {
		if (which_TD[i_node] && (mp->nodes[i_node].time >= start_period) && (mp->nodes[i_node].time <= end_period)) {
			active_nodes_indices[i_node] = nb_active_nodes;
			nb_active_nodes++;
		}
		else {
			which_TD[i_node] = false;
		}
	}

	// Acbtual model formulation starts here
	//double to_return = 0;
	solver_result to_return;
	to_return.obj_value = 0;
	to_return.optimal = false;

	try {
		GRBModel model = GRBModel (*mp->env);
		model.set(GRB_StringAttr_ModelName, "Facility planning");
		model.set(GRB_IntAttr_ModelSense, GRB_MINIMIZE);
		model.set(GRB_DoubleParam_TimeLimit, mp->TimeLimit);
		model.set(GRB_DoubleParam_MIPGap, 0.001);
		model.set(GRB_IntParam_OutputFlag, output_flag);
		model.set(GRB_StringParam_LogFile, mp->log_file);
		model.set(GRB_IntParam_Method, 2);
		model.set(GRB_DoubleParam_Heuristics, 0.75);
		model.set(GRB_IntParam_Threads, 4);
		//string log_file_name = "C:/Users/krzys/source/repos/StaticFacilityPlanning/b_" + to_string((int)b) + "_m_" + to_string((int)m_cost) + "_instance_" + to_string((int)i_instance) + "log.txt";
		//model.set(GRB_StringParam_LogFile, log_file_name.c_str());

		// Variable x
		GRBVar** x;
		x = new GRBVar * [mp->nb_locations];
		for (int i_l = 0; i_l < mp->nb_locations; i_l++) {
			x[i_l] = model.addVars(nb_active_nodes, GRB_BINARY); // Make them binary
		}

		// Variable y_0
		GRBVar*** y_0;
		y_0 = new GRBVar * *[mp->nb_customers];
		for (int i_c = 0; i_c < mp->nb_customers; i_c++) {
			y_0[i_c] = new GRBVar * [mp->nb_locations];

			for (int i_l = 0; i_l < mp->nb_locations; i_l++) {
				y_0[i_c][i_l] = model.addVars(nb_active_nodes); // make them unbounded
			}
		}

		// Variable w_0
		GRBVar** w_0;
		w_0 = new GRBVar * [mp->nb_customers];

		for (int i_c_2 = 0; i_c_2 < mp->nb_customers; i_c_2++) {
			w_0[i_c_2] = model.addVars(nb_active_nodes); // make them unbounded
		}

		// ub variable
		GRBVar* ub_cost;
		ub_cost = model.addVars(mp->nb_nodes);
		for (int i_node = 0; i_node < mp->nb_nodes; i_node++) {
			ub_cost[i_node].set(GRB_DoubleAttr_LB, -GRB_INFINITY);;
		}

		// mu variables
		int nb_robust_constraints = nb_active_nodes * (mp->nb_locations + 2 * mp->nb_customers + mp->nb_locations * mp->nb_customers + 1);

		// Creating the linear expressions to store each robust constraint
		GRBLinExpr expr_constant = 0;

		// kappa variable 
		GRBVar* kappa;
		kappa = model.addVars(mp->nb_nodes);
		for (int i_node = 0; i_node < mp->nb_nodes; i_node++) {
			kappa[i_node].set(GRB_DoubleAttr_LB, -GRB_INFINITY);
		}

		model.setObjective(kappa[mp->nb_nodes - 1] - kappa[0]);
		//model.setObjective(ub_cost[1] - kappa[0]);

		// Constraint normalizing the kappas
		model.addConstr(kappa[0] == 0);

		for (int i_arc = 0; i_arc < mp->nb_of_arcs; i_arc++) {
			if ((mp->nodes[mp->arcs[i_arc].end_node].time >= start_period) && (mp->nodes[mp->arcs[i_arc].end_node].time <= end_period)) {
				if (which_TD[mp->arcs[i_arc].end_node] && which_TD[mp->arcs[i_arc].start_node]) {
					//std::cout << "Node: " << mp->arcs[i_arc].start_node << " less than " << mp->arcs[i_arc].end_node << " with cost of " << mp->arcs[i_arc].end_node << "\n";
					model.addConstr(ub_cost[mp->arcs[i_arc].end_node] <= kappa[mp->arcs[i_arc].end_node] - kappa[mp->arcs[i_arc].start_node]);
				}
			}
			else {
				if ((mp->nodes[mp->arcs[i_arc].start_node].time == (start_period - 1)) && which_TD[mp->arcs[i_arc].end_node]) {
					model.addConstr(ub_cost[mp->arcs[i_arc].end_node] <= kappa[mp->arcs[i_arc].end_node] - kappa[mp->arcs[i_arc].start_node]);
				}

				if ((mp->nodes[mp->arcs[i_arc].end_node].time == (end_period + 1)) && which_TD[mp->arcs[i_arc].start_node]) {
					model.addConstr(kappa[mp->arcs[i_arc].end_node] - kappa[mp->arcs[i_arc].start_node] >= 0);
				}

				if ((mp->nodes[mp->arcs[i_arc].start_node].time < (start_period - 1)) || (mp->nodes[mp->arcs[i_arc].end_node].time > (end_period + 1)))
				{
					model.addConstr(kappa[mp->arcs[i_arc].end_node] - kappa[mp->arcs[i_arc].start_node] >= 0);
				}
			}
		}

		// Nondecreasingness of the facility location vectors, go over all nodes
		for (int i_node = 1; i_node < mp->nb_nodes - 1; i_node++) {
			if (which_TD[i_node] && (mp->nodes[i_node].time < mp->T - 1)) {
				for (int i_child = 0; i_child < mp->nodes[i_node].nb_of_children; i_child++) {
					if (which_TD[mp->nodes[i_node].children[i_child]]) {
						for (int i_facility = 0; i_facility < mp->nb_locations; i_facility++) {
							model.addConstr(x[i_facility][active_nodes_indices[i_node]] <= x[i_facility][active_nodes_indices[mp->nodes[i_node].children[i_child]]]);
						}
					}
				}
			}
		}

		// Now starts the real thing
		int offset_active_node = 0;
		int offset_RC = 0;
		int this_node;

		for (int i_node = 1; i_node < mp->nb_nodes - 1; i_node++) {
			if (which_TD[i_node]) {// Only active nodes

				// Making sure the beginning capacity is at least the given level
				if (mp->nodes[i_node].time == start_period) {
					for (int i_facility = 0; i_facility < mp->nb_locations; i_facility++) {
						model.addConstr(x[i_facility][offset_active_node] >= start_facilities[i_facility]);
					}
				}

				if (mp->nodes[i_node].time == end_period) {
					for (int i_facility = 0; i_facility < mp->nb_locations; i_facility++) {
						model.addConstr(x[i_facility][offset_active_node] >= end_facilities[i_facility]);
					}
				}

				double p = mp->nodes[i_node].total_demand / mp->nb_factors;

				/////////////////////////////////////////////////////////////
				if (for_lower_bound) {
					for (int i_c = 0; i_c < mp->nb_customers; i_c++) {
						double nom_demand = 0;
						for (int i_f = 0; i_f < mp->nb_factors; i_f++) {
							nom_demand += mp->Phi[i_c][i_f] * p;
						}
						mp->customers[i_c].nominal_demand = nom_demand;
					}
				} else {
					// here updating shit according to the other ones.
				}

				// Capacity constraints
				for (int i_l = 0; i_l < mp->nb_locations; i_l++) { // Zero the linear expressions
					expr_constant = -x[i_l][offset_active_node] * mp->factories[i_l].capacity;

					// Build up constraint material
					for (int i_c = 0; i_c < mp->nb_customers; i_c++) {
						expr_constant += y_0[i_c][i_l][offset_active_node];
					}

					model.addConstr(expr_constant <= 0.0);
					offset_RC++;
				}

				// Upper bounding unmet demand costs - part I, w == demand minus delivery
				for (int i_c = 0; i_c < mp->nb_customers; i_c++) {
					// Zero the linear expressions
					expr_constant = -w_0[i_c][offset_active_node];

					//expr_variable[i_c][offset_RC] += 1.0;

					// Build up constraint material
					for (int i_l = 0; i_l < mp->nb_locations; i_l++) {
						expr_constant -= y_0[i_c][i_l][offset_active_node];
					}

					expr_constant += mp->customers[i_c].nominal_demand;
					model.addConstr(expr_constant == 0);
					offset_RC++;
				}

				// Now the constraint on the total cost in a given node
				this_node = offset_RC;
				expr_constant = -ub_cost[i_node];
				//std::cout << "Maintenance cost:  " <<  mp->m_cost << " " << mp->b << "\n";

				for (int i_l = 0; i_l < mp->nb_locations; i_l++) {
					expr_constant += x[i_l][offset_active_node] * mp->m_cost;
				}

				for (int i_c = 0; i_c < mp->nb_customers; i_c++) {
					expr_constant += w_0[i_c][offset_active_node] * mp->b;

					for (int i_l = 0; i_l < mp->nb_locations; i_l++) {
						//std::cout << "Cost of transport: " << (mp->c[i_l][i_c] - mp->revenue) << "\n";
						expr_constant += y_0[i_c][i_l][offset_active_node] * (mp->c[i_l][i_c] - mp->revenue);
					}
				}

				model.addConstr(expr_constant <= 0.0);
				offset_RC++;
				offset_active_node++;
			}
			else {
				model.addConstr(ub_cost[i_node] == 0);
			}
		}

		model.update();

		if (mp->solution_method < 2) {// This part activated only if the method number is 0
			int numvars = model.get(GRB_IntAttr_NumVars);
			GRBVar* vars;
			vars = model.getVars();
			std::chrono::steady_clock::time_point start_moment;
			start_moment = std::chrono::steady_clock::now();
			std::string incumbent_filename = "./incumbent_" + std::to_string(mp->i_instance) + "_" + std::to_string(mp->solution_method) + ".txt";

			std::ofstream incumbent_file(incumbent_filename.c_str());
			if (incumbent_file.fail()) {
				std::cout << "Failed to open outputfile.\n";
			}

			mycallback_dynamic cb = mycallback_dynamic(numvars, vars, &incumbent_file,
				mp, which_TD, active_nodes_indices, nb_active_nodes, start_moment);
			model.setCallback(&cb);
			model.optimize();
			incumbent_file.close();

		} else {
			model.optimize();
		}

		if (AddPolicies) {
			// Add the new decisions to the database
			for (int i_node = 1; i_node < mp->nb_nodes - 1; i_node++) {
				if (which_TD[i_node]) {
					int t = mp->nodes[i_node].time;
					offset_active_node = active_nodes_indices[i_node];

					// Check if maybe such a decision is there already

					bool unique_or_not = true;

					for (int i_policy_to_compare = 0; i_policy_to_compare < mp->how_many_dec_there[t]; i_policy_to_compare++) {
						int the_same = 0;
						for (int i_f = 0; i_f < mp->nb_locations; i_f++) {
							if (abs(x[i_f][offset_active_node].get(GRB_DoubleAttr_X) - mp->policies_x[t][i_policy_to_compare][i_f]) < 0.1) {
								the_same++;
							}
						}

						if (the_same > mp->nb_locations - 1) {
							unique_or_not = false;
							break;
						}
					}

					if (unique_or_not) {
						for (int i_facility = 0; i_facility < mp->nb_locations; i_facility++) {
							mp->policies_x[t][mp->how_many_dec_there[t]][i_facility] = abs(round(x[i_facility][offset_active_node].get(GRB_DoubleAttr_X)));
						}
						mp->how_many_dec_there[t]++;
					}
				}
			}
		}

		if (model.get(GRB_IntAttr_Status) == 9)
			to_return.obj_value = model.get(GRB_DoubleAttr_ObjVal);  //kappa[mp->nb_nodes - 1].get(GRB_DoubleAttr_X);
			to_return.lower_bound = model.get(GRB_DoubleAttr_ObjBound);
		else {
			if (model.get(GRB_IntAttr_Status) == 2) {
				to_return.obj_value = model.get(GRB_DoubleAttr_ObjVal);
				to_return.lower_bound = model.get(GRB_DoubleAttr_ObjVal);
				to_return.optimal = true;
			}
			else {
				to_return.obj_value = 10000000;
				to_return.lower_bound = 0;
			}
		}

		delete[] x;
		delete[] y_0;
		delete[] w_0;
		delete[] ub_cost;
		delete[] kappa;
	}
	catch (GRBException e) {
		cout << "Error code = " << e.getErrorCode() << endl;
		cout << e.getMessage() << endl;
	}
	catch (...) {
		cout << "Exception during optimization" << endl;
	}

	return to_return;
}

void count_unique_policies_to_array(problem_state* mp, double** unique_policies, int round_index) {
	for (int t = 0; t < mp->T; t++) {
		//std::cout << "How many dec: " << mp->how_many_dec_there[t] << "\n";

		mp->how_many_unique_dec_there[t] = mp->how_many_dec_there[t];
	}

	for (int t = 0; t < mp->T; t++) {
		unique_policies[round_index][t] = mp->how_many_unique_dec_there[t];
	}
}

void export_policies(problem_state* mp) {
	std::ofstream my_file;
	my_file.open(mp->policies_file.c_str());

	for (int t = 0; t < mp->T; t++) {
		my_file << mp->how_many_dec_there[t] << " ";
		my_file.flush();
	}

	//my_file << "\n";//TODO

	for (int t = 0; t < mp->T; t++) {
		for (int i_policy = 0; i_policy < mp->how_many_dec_there[t]; i_policy++) {
			for (int i_l = 0; i_l < mp->nb_locations; i_l++) {
				my_file << mp->policies_x[t][i_policy][i_l] << " ";
				my_file.flush();
			}
			//my_file << "\n"; //TODO
		}
		//my_file << "\n";//TODO
	}

	my_file.close();
}

void import_policies(problem_state* mp) {
	std::fstream my_file;
	my_file.open(mp->policies_file.c_str());

	for (int t = 0; t < mp->T; t++) {
		my_file >> mp->how_many_dec_there[t];
		std::cout << mp->how_many_dec_there[t] << " ";
	}

	for (int t = 0; t < mp->T; t++) {
		for (int i_policy = 0; i_policy < mp->how_many_dec_there[t]; i_policy++) {
			for (int i_l = 0; i_l < mp->nb_locations; i_l++) {
				my_file >> mp->policies_x[t][i_policy][i_l];
				std::cout << mp->policies_x[t][i_policy][i_l] << " ";
			}
		}
	}

	my_file.close();
}

double dynamic_solution_CCG(bool* which_TD, int start_period, int end_period, double* start_facilities, double* end_facilities, problem_state* mp, int output_flag, bool AddPolicies) {
	std::cout << "============= Solving the dynamic problem with CCG =========\n";
	int nb_active_nodes = 0;
	int* active_nodes_indices = new int[mp->nb_nodes];
	std::cout << "Backlogging: " << mp->b << " maintenance: " << mp->m_cost << " revenue: " << mp->revenue << "\n";

	for (int i_node = 1; i_node < mp->nb_nodes - 1; i_node++) {
		if (which_TD[i_node] && (mp->nodes[i_node].time >= start_period) && (mp->nodes[i_node].time <= end_period)) {
			active_nodes_indices[i_node] = nb_active_nodes;
			nb_active_nodes++;
		}
		else {
			which_TD[i_node] = false;
		}
	}

	//double to_return = 0;
	//bool to_return_optimal = false;

	solver_result to_return;
	to_return.obj_value = 0;
	to_return.optimal = false;

	try {
		// Acbtual model formulation starts here
		GRBModel model = GRBModel (*mp->env);
		model.set(GRB_StringAttr_ModelName, "Facility planning");
		model.set(GRB_IntAttr_ModelSense, GRB_MINIMIZE);
		model.set(GRB_DoubleParam_TimeLimit, mp->TimeLimit);
		model.set(GRB_DoubleParam_MIPGap, 0.001);
		model.set(GRB_IntParam_OutputFlag, output_flag);
		model.set(GRB_StringParam_LogFile, mp->log_file);
		model.set(GRB_IntParam_Method, 2);
		model.set(GRB_DoubleParam_Heuristics, 0.75);
		model.set(GRB_IntParam_Threads, 4);
		const int nb_vertices = 6;

		double vertices[6][4] = {
			{1 - mp->uncertainty_level, 1 - mp->uncertainty_level, 1 + mp->uncertainty_level, 1 + mp->uncertainty_level},
			{1 - mp->uncertainty_level, 1 + mp->uncertainty_level, 1 - mp->uncertainty_level, 1 + mp->uncertainty_level},
			{1 + mp->uncertainty_level, 1 - mp->uncertainty_level, 1 - mp->uncertainty_level, 1 + mp->uncertainty_level},
			{1 - mp->uncertainty_level, 1 + mp->uncertainty_level, 1 + mp->uncertainty_level, 1 - mp->uncertainty_level},
			{1 + mp->uncertainty_level, 1 - mp->uncertainty_level, 1 + mp->uncertainty_level, 1 - mp->uncertainty_level},
			{1 + mp->uncertainty_level, 1 + mp->uncertainty_level, 1 - mp->uncertainty_level, 1 - mp->uncertainty_level},
		};

		// Variable x
		GRBVar** x;
		x = new GRBVar * [mp->nb_locations];
		for (int i_l = 0; i_l < mp->nb_locations; i_l++) {
			x[i_l] = model.addVars(nb_active_nodes, GRB_BINARY); // Make them binary
		}

		// Variable y_0
		GRBVar**** y_0;
		y_0 = new GRBVar ***[mp->nb_customers];
		for (int i_c = 0; i_c < mp->nb_customers; i_c++) {
			y_0[i_c] = new GRBVar ** [mp->nb_locations];

			for (int i_l = 0; i_l < mp->nb_locations; i_l++) {
				y_0[i_c][i_l] = new GRBVar * [nb_active_nodes];

				for (int i_node = 0; i_node < nb_active_nodes; i_node++) {
					y_0[i_c][i_l][i_node] = model.addVars(nb_vertices, GRB_CONTINUOUS); // make them unbounded
				}
			}
		}

		// Variable w_0
		GRBVar*** w_0;
		w_0 = new GRBVar ** [mp->nb_customers];

		for (int i_c_2 = 0; i_c_2 < mp->nb_customers; i_c_2++) {
			w_0[i_c_2] = new GRBVar * [nb_active_nodes];

			for (int i_node = 0; i_node < nb_active_nodes; i_node++) {
				w_0[i_c_2][i_node] = model.addVars(nb_vertices, GRB_CONTINUOUS); // make them unbounded
			}
		}

		// ub variable
		GRBVar* ub_cost;
		ub_cost = model.addVars(mp->nb_nodes, GRB_CONTINUOUS);
		for (int i_node = 0; i_node < mp->nb_nodes; i_node++) {
			ub_cost[i_node].set(GRB_DoubleAttr_LB, -GRB_INFINITY);
		}

		// Creating the linear expressions to store each robust constraint
		GRBLinExpr expr_constant = 0;

		// kappa variable 
		GRBVar* kappa;
		kappa = model.addVars(mp->nb_nodes, GRB_CONTINUOUS);
		for (int i_node = 0; i_node < mp->nb_nodes; i_node++) {
			kappa[i_node].set(GRB_DoubleAttr_LB, -GRB_INFINITY);
		}

		model.setObjective(kappa[mp->nb_nodes - 1] - kappa[0]);

		// Constraint normalizing the kappas
		model.addConstr(kappa[0] == 0);

		for (int i_arc = 0; i_arc < mp->nb_of_arcs; i_arc++) {
			if ((mp->nodes[mp->arcs[i_arc].end_node].time >= start_period) && (mp->nodes[mp->arcs[i_arc].end_node].time <= end_period)) {
				if (which_TD[mp->arcs[i_arc].end_node] && which_TD[mp->arcs[i_arc].start_node]) {
					//std::cout << "Node: " << mp->arcs[i_arc].start_node << " less than " << mp->arcs[i_arc].end_node << " with cost of " << mp->arcs[i_arc].end_node << "\n";
					//std::cout << "Arc " << i_arc << "\n";
					model.addConstr(ub_cost[mp->arcs[i_arc].end_node] <= kappa[mp->arcs[i_arc].end_node] - kappa[mp->arcs[i_arc].start_node]);
				}
			}
			else {
				if ((mp->nodes[mp->arcs[i_arc].start_node].time == (start_period - 1)) && which_TD[mp->arcs[i_arc].end_node]) {
					model.addConstr(ub_cost[mp->arcs[i_arc].end_node] <= kappa[mp->arcs[i_arc].end_node] - kappa[mp->arcs[i_arc].start_node]);
				}

				if ((mp->nodes[mp->arcs[i_arc].end_node].time == (end_period + 1)) && which_TD[mp->arcs[i_arc].start_node]) {
					model.addConstr(kappa[mp->arcs[i_arc].end_node] - kappa[mp->arcs[i_arc].start_node] >= 0);
				}

				if ((mp->nodes[mp->arcs[i_arc].start_node].time < (start_period - 1)) || (mp->nodes[mp->arcs[i_arc].end_node].time > (end_period + 1)))
				{
					model.addConstr(kappa[mp->arcs[i_arc].end_node] - kappa[mp->arcs[i_arc].start_node] >= 0);
				}
			}
		}

		//std::cout << "Past arcs constraints\n";

		// Nondecreasingness of the facility location vectors, go over all nodes
		for (int i_node = 1; i_node < mp->nb_nodes - 1; i_node++) {
			if (which_TD[i_node] && (mp->nodes[i_node].time < mp->T - 1)) {
				for (int i_child = 0; i_child < mp->nodes[i_node].nb_of_children; i_child++) {
					if (which_TD[mp->nodes[i_node].children[i_child]]) {
						for (int i_facility = 0; i_facility < mp->nb_locations; i_facility++) {
							model.addConstr(x[i_facility][active_nodes_indices[i_node]] <= x[i_facility][active_nodes_indices[mp->nodes[i_node].children[i_child]]]);
						}
					}
				}
			}
		}

		// Now starts the real thing
		double demand_in_vertex_customer;
		int offset_active_node = 0;

		for (int i_node = 1; i_node < mp->nb_nodes - 1; i_node++) {
			if (which_TD[i_node]) {// Only active nodes
				// Making sure the beginning capacity is at least the given level
				if (mp->nodes[i_node].time == start_period) {
					for (int i_facility = 0; i_facility < mp->nb_locations; i_facility++) {
						model.addConstr(x[i_facility][offset_active_node] >= start_facilities[i_facility]);
					}
				}

				if (mp->nodes[i_node].time == end_period) {
					for (int i_facility = 0; i_facility < mp->nb_locations; i_facility++) {
						model.addConstr(x[i_facility][offset_active_node] >= end_facilities[i_facility]);
					}
				}

				// Capacity constraints
				for (int i_vertex = 0; i_vertex < nb_vertices; i_vertex++){
					for (int i_l = 0; i_l < mp->nb_locations; i_l++) {
						expr_constant = - x[i_l][offset_active_node] * mp->factories[i_l].capacity;

						for (int i_c = 0; i_c < mp->nb_customers; i_c++) {
							expr_constant += y_0[i_c][i_l][offset_active_node][i_vertex];
						}

						model.addConstr(expr_constant <= 0.0);
					}
				}

				double p = mp->nodes[i_node].total_demand;

				// Upper bounding unmet demand costs - part I, w == demand minus delivery
				for (int i_vertex = 0; i_vertex < nb_vertices; i_vertex++){
					for (int i_c = 0; i_c < mp->nb_customers; i_c++) { //Computing the demand of a given customer in a given vertex
						demand_in_vertex_customer = 0;

						for (int i_f = 0; i_f < mp->nb_factors; i_f++) {
							demand_in_vertex_customer += mp->Phi[i_c][i_f] * (p / mp->nb_factors) * vertices[i_vertex][i_f];
						}

						expr_constant = - w_0[i_c][offset_active_node][i_vertex] + demand_in_vertex_customer;

						// Build up constraint material
						for (int i_l = 0; i_l < mp->nb_locations; i_l++) {
							expr_constant -= y_0[i_c][i_l][offset_active_node][i_vertex];
						}

						model.addConstr(expr_constant == 0.0);
					}
				}

				// Now the constraint on the total cost in a given node
				for (int i_vertex = 0; i_vertex < nb_vertices; i_vertex++){
					expr_constant = -ub_cost[i_node]; //std::cout << "Maintenance cost:  " <<  mp->m_cost << " " << mp->b << "\n";

					for (int i_l = 0; i_l < mp->nb_locations; i_l++) {
						expr_constant += x[i_l][offset_active_node] * mp->m_cost;
					}

					for (int i_c = 0; i_c < mp->nb_customers; i_c++) {
						expr_constant += w_0[i_c][offset_active_node][i_vertex] * mp->b;

						for (int i_l = 0; i_l < mp->nb_locations; i_l++) {
							expr_constant += y_0[i_c][i_l][offset_active_node][i_vertex] * (mp->c[i_l][i_c] - mp->revenue);
						}
					}

					model.addConstr(expr_constant <= 0.0);
				}

				offset_active_node++;
			}
			else {
				model.addConstr(ub_cost[i_node] == 0);
			}
		}

		model.update();

		if (mp->solution_method < 2) {// This part activated only if the method number is 0
			int numvars = model.get(GRB_IntAttr_NumVars);
			GRBVar* vars;
			vars = model.getVars();
			std::chrono::steady_clock::time_point start_moment;
			start_moment = std::chrono::steady_clock::now();
			std::string incumbent_filename = "./incumbent_" + std::to_string(mp->i_instance) + "_" + std::to_string(mp->solution_method) + ".txt";

			std::ofstream incumbent_file(incumbent_filename.c_str());
			if (incumbent_file.fail()) {
				std::cout << "Failed to open outputfile.\n";
			}

			mycallback_dynamic cb = mycallback_dynamic(numvars, vars, &incumbent_file,
				mp, which_TD, active_nodes_indices, nb_active_nodes, start_moment);
			model.setCallback(&cb);
			model.optimize();
			incumbent_file.close();

		} else {
			model.optimize();
		}

		if (AddPolicies) {
			// Add the new decisions to the database
			for (int i_node = 1; i_node < mp->nb_nodes - 1; i_node++) {
				if (which_TD[i_node]) {
					int t = mp->nodes[i_node].time;
					offset_active_node = active_nodes_indices[i_node];

					// Check if maybe such a decision is there already
					bool unique_or_not = true;

					for (int i_policy_to_compare = 0; i_policy_to_compare < mp->how_many_dec_there[t]; i_policy_to_compare++) {
						int the_same = 0;
						for (int i_f = 0; i_f < mp->nb_locations; i_f++) {
							if (abs(x[i_f][offset_active_node].get(GRB_DoubleAttr_X) - mp->policies_x[t][i_policy_to_compare][i_f]) < 0.1) {
								the_same++;
							}
						}

						if (the_same > mp->nb_locations - 1) {
							unique_or_not = false;
							break;
						}
					}

					if (unique_or_not) {
						for (int i_facility = 0; i_facility < mp->nb_locations; i_facility++) {
							mp->policies_x[t][mp->how_many_dec_there[t]][i_facility] = abs(round(x[i_facility][offset_active_node].get(GRB_DoubleAttr_X)));
						}
						mp->how_many_dec_there[t]++;
					}
				}
			}
		}

		//////////////////////////////////////
		// Determining the worst-case vertex realizations

		for (int i_node = 1; i_node < mp->nb_nodes - 1; i_node++) {
			if (which_TD[i_node]) {// Only active nodes
				// Condition if the total costs in this vertex equal to the node totals
				for (int i_vertex = 0; i_vertex < nb_vertices; i_vertex++){
					double max_minus_node = - ub_cost[i_node].get(GRB_DoubleAttr_X);

					for (int i_l = 0; i_l < mp->nb_locations; i_l++) {
						max_minus_node += x[i_l][offset_active_node].get(GRB_DoubleAttr_X) * mp->m_cost;
					}

					for (int i_c = 0; i_c < mp->nb_customers; i_c++) {
						max_minus_node += w_0[i_c][offset_active_node][i_vertex].get(GRB_DoubleAttr_X) * mp->b;

						for (int i_l = 0; i_l < mp->nb_locations; i_l++) {
							max_minus_node += y_0[i_c][i_l][offset_active_node][i_vertex].get(GRB_DoubleAttr_X) * (mp->c[i_l][i_c] - mp->revenue);
						}
					}

					if (max_minus_node >= -0.001) {// if a given vertex is tight
						which_TD_which_vertex[i_node] = i_vertex;
					}
				}

				offset_active_node++;
			}
		}

		//////////////////////////////////////


		if (model.get(GRB_IntAttr_Status) == 9)
			to_return.obj_value = model.get(GRB_DoubleAttr_ObjVal);  //kappa[mp->nb_nodes - 1].get(GRB_DoubleAttr_X);
			to_return.lower_bound = model.get(GRB_DoubleAttr_ObjBound);
		else {
			if (model.get(GRB_IntAttr_Status) == 2) {
				to_return.optimal = true;
				to_return.obj_value = model.get(GRB_DoubleAttr_ObjVal);
				to_return.lower_bound = model.get(GRB_DoubleAttr_ObjVal);
			}
			else {
				to_return.obj_value = 1000000;
				to_return.lower_bound = 0;
			}
		}

		delete[] x;
		delete[] y_0;
		delete[] w_0;
		delete[] ub_cost;
		delete[] kappa;
		//delete[] demand_in_vertices;
	}
	catch (GRBException e) {
		cout << "Error code = " << e.getErrorCode() << endl;
		cout << e.getMessage() << endl;
	}
	catch (...) {
		cout << "Exception during optimization" << endl;
	}

	return to_return;
}
