#include <stdlib.h>
#include <string>
#include <cmath>
#include <time.h>
#include <fstream>
#include <iostream>
#include <ctime>
#include <sstream>
#include "gurobi_c++.h"
#include <sstream>
#include <cstdlib>
#include "structures_header.h"
#include "func_header_modified.h"
#include <omp.h>
#include <chrono>
//#include "func_header_modified.h"
using namespace std;

int main(int argc, char* argv[]) {
	const int nb_locations = atoi(argv[1]);
	const int nb_customers = 2 * nb_locations; // TODO
	const int T = atoi(argv[2]);// TODO
	double routing_cost = 1;
	double b = 30;
	double revenue = 80;
	int i_instance = atoi(argv[3]);
	int method_used = 2;
	int improvement_rounds = 1;
	double max_method_time = 5 * 3600;
	int how_many_sorted = 3;
	const int max_dec_per_period = (improvement_rounds + 1) * (T + 1);
	double demand_step_size = 50;
	double base_demand = 100;
	double uncertainty_level = 0.4;
	double facility_cost_factor = 0.8;
	double capacity = 1.5 * (base_demand + T * demand_step_size) / nb_locations;
	double m_cost = capacity * revenue * facility_cost_factor;
	std::string log_filename = "./log_file_" + std::to_string(nb_locations) + "_" + std::to_string(T) + "_" + std::to_string(i_instance) + "_" + std::to_string(method_used) + ".txt";
	std::string results_filename = "./results_" + std::to_string(nb_locations) + "_" + std::to_string(T) + "_" + std::to_string(i_instance) + "_" + std::to_string(method_used) + ".txt";
	std::string time_obj_filename = "./time_obj_" + std::to_string(nb_locations) + "_" + std::to_string(T) + "_" + std::to_string(i_instance) + "_" + std::to_string(method_used) + ".txt";
	std::string policies_filename = "./policies_" + std::to_string(nb_locations) + "_" + std::to_string(T) + "_" + std::to_string(i_instance) + ".txt";

	std::ofstream myfile(results_filename.c_str());
	if (myfile.fail()){
		std::cout << "Failed to open outputfile.\n";
	}

	std::ofstream time_obj_file(time_obj_filename.c_str());
	if (time_obj_file.fail()) {
		std::cout << "Failed to open time obj file.\n";
	}

	//myfile.open(results_filename.c_str());// , std::ios::app);

	// Building the list of time x demand_regime nodes
	int nb_nodes = (T + 1) * (T + 2) / 2 + 1;
	t_dr_node* nodes;
	nodes = new t_dr_node[nb_nodes];

	nodes[0].time = -1;
	nodes[0].total_demand = 0;
	nodes[0].i_dr = 0;
	nodes[nb_nodes - 1].time = T;
	nodes[nb_nodes - 1].total_demand = 0;
	nodes[nb_nodes - 1].i_dr = 0;
	int i_node = 1;

	for (int t = 0; t < T; t++)
	{
		for (int i_dr = 0; i_dr < t + 2; i_dr++)
		{
			nodes[i_node].time = t;
			nodes[i_node].i_dr = i_dr;
			nodes[i_node].total_demand = base_demand + (nodes[i_node].i_dr) * demand_step_size;
			//std::cout << "TB node: " << i_node << " time: " << nodes[i_node].time << " i_dr: " << nodes[i_node].i_dr << " total demand: " << nodes[i_node].total_demand << "\n" ;
			i_node++;
		}
	}

	// Now identifying the children and parents of each node
	int** map_TB_to_node;
	map_TB_to_node = new int* [T];
	int* how_many_i_dr;
	how_many_i_dr = new int[T];

	for (int t = 0; t < T; t++)
	{
		how_many_i_dr[t] = t + 2;
		map_TB_to_node[t] = new int[how_many_i_dr[t]]; // matrix that will map a combination of Time x Demand regime to a node in our list
	}

	int nb_of_arcs = 0;
	for (int i_node = 0; i_node < nb_nodes; i_node++)
	{
		if ((i_node > 0) && (i_node < nb_nodes - 1))
		{
			map_TB_to_node[nodes[i_node].time][nodes[i_node].i_dr] = i_node;
		}
	}

	for (int i_node = 0; i_node < nb_nodes; i_node++)
	{
		//std::cout << "i node: " << i_node << "\n";
		if (nodes[i_node].time < T - 1)
		{			// two children
			nodes[i_node].nb_of_children = 2;
			nodes[i_node].children = new int[nodes[i_node].nb_of_children];
			nodes[i_node].children[0] = map_TB_to_node[nodes[i_node].time + 1][nodes[i_node].i_dr];
			nodes[i_node].children[1] = map_TB_to_node[nodes[i_node].time + 1][nodes[i_node].i_dr + 1];
			//std::cout << "Node: " << i_node << " " << nodes[i_node].children[0] << " " << nodes[i_node].children[1] << "\n";
			nb_of_arcs += 2;
		}
		else if (nodes[i_node].time == T - 1)
		{
			// one child
			nodes[i_node].nb_of_children = 1;
			nodes[i_node].children = new int[nodes[i_node].nb_of_children];
			nodes[i_node].children[0] = nb_nodes - 1;
			nb_of_arcs += 1;
			//std::cout << "Node: " << i_node << " " << nodes[i_node].children[0] << "\n";
		}
		else
		{
			nodes[i_node].nb_of_children = 0;
			// no children
		}

		// Print all children
		//std::cout << "Node: " << i_node << " children: ";
		for (int i_child = 0; i_child < nodes[i_node].nb_of_children; i_child++)
		{
			//std::cout << nodes[i_node].children[i_child] << " ";
		}
		//std::cout << "\n";

		if ((nodes[i_node].time > -1) && (nodes[i_node].time < T))
		{
			if (nodes[i_node].time == 0) {
				nodes[i_node].nb_of_parents = 1;
				nodes[i_node].parents = new int[nodes[i_node].nb_of_parents];
				nodes[i_node].parents[0] = 0;
			}
			else {
				if (nodes[i_node].time < nodes[i_node].i_dr) {
					nodes[i_node].nb_of_parents = 1;
					nodes[i_node].parents = new int[nodes[i_node].nb_of_parents];
					nodes[i_node].parents[0] = map_TB_to_node[nodes[i_node].time - 1][nodes[i_node].i_dr - 1];
					//std::cout << "Node: " << i_node << " parents: " << parents[i_node][0] << "\n";
				}
				else if (nodes[i_node].i_dr > 0) {
					nodes[i_node].nb_of_parents = 2;
					nodes[i_node].parents = new int[nodes[i_node].nb_of_parents];
					nodes[i_node].parents[0] = map_TB_to_node[nodes[i_node].time - 1][nodes[i_node].i_dr];
					nodes[i_node].parents[1] = map_TB_to_node[nodes[i_node].time - 1][nodes[i_node].i_dr - 1];
					//std::cout << "Node: " << i_node << " parents: " << parents[i_node][0] << " " << parents[i_node][1] << "\n";
				}
				else {
					nodes[i_node].nb_of_parents = 1;
					nodes[i_node].parents = new int[nodes[i_node].nb_of_parents];
					nodes[i_node].parents[0] = map_TB_to_node[nodes[i_node].time - 1][nodes[i_node].i_dr];
					//std::cout << "Node: " << i_node << " parents: " << parents[i_node][0] << "\n";
				}
			}
		}
		else
		{
			nodes[i_node].nb_of_parents = 0;
		}

		// Print all children
		//std::cout << "Node: " << i_node << " parents: ";
		for (int i_parent = 0; i_parent < nodes[i_node].nb_of_parents; i_parent++)
		{
			//std::cout << nodes[i_node].parents[i_parent] << " ";
		}
		//std::cout << "\n";
	}

	int offset = 0;
	arc* arcs;
	arcs = new arc[nb_of_arcs];

	for (int i_node = 0; i_node < nb_nodes; i_node++)
	{
		for (int j = 0; j < nodes[i_node].nb_of_children; j++)
		{
			arcs[offset].start_node = i_node;
			arcs[offset].end_node = nodes[i_node].children[j];
			//std::cout << "Offset= " << offset << " " << arcs_data[offset][0] << " " << arcs_data[offset][1] << "\n";
			offset++;
		}
	}

	for (int i_node = 1; i_node < nb_nodes; i_node++)
	{
		//std::cout << "\nNode " << i_node << "\n";
		for (int i_child = 0; i_child < nodes[i_node].nb_of_children; i_child++)
		{
			//std::cout << nodes[i_node].children[i_child] << " ";
		}
	}

	srand(i_instance);
	// Sampling the matrix with factories' coordinates
	factory* factories;
	factories = new factory[nb_locations];

	for (int i_f = 0; i_f < nb_locations; i_f++) {
		factories[i_f].x = ((double)rand() / (RAND_MAX)) * 10;
		factories[i_f].y = ((double)rand() / (RAND_MAX)) * 10;
		factories[i_f].capacity = capacity;
		//std::cout << "Capacity: " << factories[i_f].capacity;
	}

	// Sampling the matrix with customers' coordinates

	customer* customers;
	customers = new customer[nb_customers];
	for (int i_c = 0; i_c < nb_customers; i_c++)
	{
		customers[i_c].x = ((double)rand() / (RAND_MAX)) * 10;
		customers[i_c].y = ((double)rand() / (RAND_MAX)) * 10;
		customers[i_c].lb_demand = 0.0;
		//std::cout << "Customer: " << i_c << " " <<  customers[i_c].x << " " << customers[i_c].y << "\n";
	}

	// Factor centers
	int nb_factors = 4;
	demand_factor* demand_factors;
	demand_factors = new demand_factor[nb_factors];

	demand_factors[0].x = 0;
	demand_factors[0].y = 0;

	demand_factors[1].x = 0;
	demand_factors[1].y = 10;

	demand_factors[2].x = 10;
	demand_factors[2].y = 0;

	demand_factors[3].x = 10;
	demand_factors[3].y = 10;

	// Transport costs
	double** c;
	c = new double* [nb_locations];
	for (int i = 0; i < nb_locations; i++) {
		c[i] = new double[nb_customers];
		for (int j = 0; j < nb_customers; j++) {
			c[i][j] = routing_cost * sqrt(pow(customers[j].x - factories[i].x, 2) + pow(customers[j].y - factories[i].y, 2));
			//std::cout << "Distance" << i << " " << j << " " << customers_coords[j][0] << " " << factories_coords[i][0] << " " << customers_coords[j][1] << " " << factories_coords[i][1] << " " << c[i][j] << "\n";
		}
	}

	// Factor sensitivities
	double** Phi;
	Phi = new double* [nb_customers];
	for (int i_c = 0; i_c < nb_customers; i_c++) {
		Phi[i_c] = new double[nb_factors];
	}

	for (int i_factor = 0; i_factor < nb_factors; i_factor++) {
		double total_distance = 0;
		for (int i_c = 0; i_c < nb_customers; i_c++) {
			total_distance += 1 / sqrt(pow(customers[i_c].x - demand_factors[i_factor].x, 2) + pow(customers[i_c].y - demand_factors[i_factor].y, 2));
		}

		for (int i_c = 0; i_c < nb_customers; i_c++) {
			Phi[i_c][i_factor] = (1 / sqrt(pow(customers[i_c].x - demand_factors[i_factor].x, 2) + pow(customers[i_c].y - demand_factors[i_factor].y, 2))) / total_distance;
			//std::cout << Phi[i_c][i_factor] << " ";
		}

		//std::cout << std::endl;
	}

	// The time x demand regime x decision nodes
	int nb_TBD_nodes = 2 + (nb_nodes - 2) * max_dec_per_period;
	t_dr_dec_node* TBD_nodes;
	TBD_nodes = new t_dr_dec_node[nb_TBD_nodes];
	TBD_nodes[0].time = -1; // Defining stuff for the first node
	TBD_nodes[0].total_demand = 0;
	TBD_nodes[0].i_dr = 0;
	TBD_nodes[0].han_cost_computed = true;
	TBD_nodes[0].total_cost_computed = false;
	TBD_nodes[0].han_cost_computed_no_IU = true;
	TBD_nodes[0].total_cost_computed_no_IU = false;
	TBD_nodes[0].han_cost = 0;
	TBD_nodes[0].han_cost_no_IU = 0;
	TBD_nodes[0].dec_number = 0;
	TBD_nodes[nb_TBD_nodes - 1].time = T; // Defining stuff for the first node

	offset = 1;
	int*** map_TBD_to_node = new int** [T];
	for (int t = 0; t < T; t++) {
		map_TBD_to_node[t] = new int* [how_many_i_dr[t]];
		for (int i_dr = 0; i_dr < t + 2; i_dr++) {
			map_TBD_to_node[t][i_dr] = new int[max_dec_per_period];
			for (int i_dec = 0; i_dec < max_dec_per_period; i_dec++) {
				map_TBD_to_node[t][i_dr][i_dec] = offset;
				offset++;
			}
		}
	}

	// Initializing the data of the TRD nodes
	offset = 1;
	for (int t = 0; t < T; t++){
		for (int i_dr = 0; i_dr < how_many_i_dr[t]; i_dr++){
			for (int i_dec = 0; i_dec < max_dec_per_period; i_dec++){
				TBD_nodes[offset].time = t;
				TBD_nodes[offset].total_demand = nodes[map_TB_to_node[t][i_dr]].total_demand;
				TBD_nodes[offset].i_dr = i_dr;
				TBD_nodes[offset].dec_number = i_dec;
				TBD_nodes[offset].han_cost_computed = false;
				TBD_nodes[offset].han_cost_computed_no_IU = false;
				TBD_nodes[offset].total_cost_computed = false;
				TBD_nodes[offset].total_cost_computed_no_IU = false;
				//std::cout << "TBD_node: " << offset << " T= " << TBD_nodes[offset].time << " B= " << TBD_nodes[offset].i_dr << " i_dec= " << TBD_nodes[offset].dec_number << "\n";

				// Building up the children list
				int TD_node = map_TB_to_node[TBD_nodes[offset].time][i_dr];
				if (t < T - 1)
				{
					TBD_nodes[offset].nb_of_children = nodes[TD_node].nb_of_children * max_dec_per_period;

					if (TBD_nodes[offset].nb_of_children > 0)
					{
						TBD_nodes[offset].children = new int[TBD_nodes[offset].nb_of_children];
						TBD_nodes[offset].children_possible = new bool[TBD_nodes[offset].nb_of_children];
						TBD_nodes[offset].children_worst = new bool[TBD_nodes[offset].nb_of_children];
						TBD_nodes[offset].children_worst_no_IU = new bool[TBD_nodes[offset].nb_of_children];
						int i_child = 0;

						for (int i_child_master = 0; i_child_master < nodes[TD_node].nb_of_children; i_child_master++)
						{
							int time_of_child = nodes[nodes[TD_node].children[i_child_master]].time;
							int i_dr_of_child = nodes[nodes[TD_node].children[i_child_master]].i_dr;

							for (int i_child_sub = 0; i_child_sub < max_dec_per_period; i_child_sub++)
							{
								//std::cout << TD_node << " " << nodes[TD_node].children[i_child_master] << " " << time_of_child << " " << i_dr_of_child << " " << i_child_sub << "\n";
								TBD_nodes[offset].children[i_child] = map_TBD_to_node[time_of_child][i_dr_of_child][i_child_sub];
								TBD_nodes[offset].children_possible[i_child] = false;
								TBD_nodes[offset].children_worst[i_child] = false;
								TBD_nodes[offset].children_worst_no_IU[i_child] = false;
								i_child++;
							}
						}
					}
				}
				else
				{
					TBD_nodes[offset].nb_of_children = 0;
				}

				// Identifying the parents list
				if (t > 1){
					TBD_nodes[offset].nb_of_parents = nodes[TD_node].nb_of_parents * max_dec_per_period;
					if (TBD_nodes[offset].nb_of_parents > 0){
						TBD_nodes[offset].parents = new int[TBD_nodes[offset].nb_of_parents];
						TBD_nodes[offset].parents_possible = new bool[TBD_nodes[offset].nb_of_parents];
						int i_parent = 0;
						for (int i_parent_master = 0; i_parent_master < nodes[TD_node].nb_of_parents; i_parent_master++){
							int time_of_parent = nodes[nodes[TD_node].parents[i_parent_master]].time;
							int i_dr_of_parent = nodes[nodes[TD_node].parents[i_parent_master]].i_dr;

							for (int i_parent_sub = 0; i_parent_sub < max_dec_per_period; i_parent_sub++){
								TBD_nodes[offset].parents[i_parent] = map_TBD_to_node[time_of_parent][i_dr_of_parent][i_parent_sub];
								TBD_nodes[offset].parents_possible[i_parent] = true;
								i_parent++;
							}
						}
					}
				} else {
					TBD_nodes[offset].nb_of_parents = 1;
					TBD_nodes[offset].parents = new int[TBD_nodes[offset].nb_of_parents];
					TBD_nodes[offset].parents[0] = 0;
				}

				// Building up the parents list
				offset++;
			}
		}
	}

	// Initializing the data of the zero TBD node
	// Building up the children list
	offset = 0;
	int TD_node = 0;
	TBD_nodes[offset].nb_of_children = nodes[TD_node].nb_of_children * max_dec_per_period;

	if (TBD_nodes[offset].nb_of_children > 0){
		TBD_nodes[offset].children = new int[TBD_nodes[offset].nb_of_children];
		TBD_nodes[offset].children_possible = new bool[TBD_nodes[offset].nb_of_children];
		TBD_nodes[offset].children_worst = new bool[TBD_nodes[offset].nb_of_children];
		TBD_nodes[offset].children_worst_no_IU = new bool[TBD_nodes[offset].nb_of_children];
		int i_child = 0;

		for (int i_child_master = 0; i_child_master < nodes[TD_node].nb_of_children; i_child_master++){
			int time_of_child = nodes[nodes[TD_node].children[i_child_master]].time;
			int i_dr_of_child = nodes[nodes[TD_node].children[i_child_master]].i_dr;

			for (int i_child_sub = 0; i_child_sub < max_dec_per_period; i_child_sub++){
				//std::cout << TD_node << " " << nodes[TD_node].children[i_child_master] << " " << time_of_child << " " << i_dr_of_child << " " << i_child_sub << "\n";
				TBD_nodes[offset].children[i_child] = map_TBD_to_node[time_of_child][i_dr_of_child][i_child_sub];
				TBD_nodes[offset].children_possible[i_child] = true;
				TBD_nodes[offset].children_worst[i_child] = true;
				TBD_nodes[offset].children_worst_no_IU[i_child] = true;
				i_child++;
			}
		}
	}

	// Building up the parents list
	TBD_nodes[offset].nb_of_parents = nodes[TD_node].nb_of_parents * max_dec_per_period;

	if (TBD_nodes[offset].nb_of_parents > 0){
		TBD_nodes[offset].parents = new int[TBD_nodes[offset].nb_of_parents];
		TBD_nodes[offset].parents_possible = new bool[TBD_nodes[offset].nb_of_parents];
		int i_parent = 0;

		for (int i_parent_master = 0; i_parent_master < nodes[TD_node].nb_of_parents; i_parent_master++){
			int time_of_parent = nodes[nodes[TD_node].parents[i_parent_master]].time;
			int i_dr_of_parent = nodes[nodes[TD_node].parents[i_parent_master]].i_dr;

			for (int i_parent_sub = 0; i_parent_sub < max_dec_per_period; i_parent_sub++)
			{
				TBD_nodes[offset].parents[i_parent] = map_TBD_to_node[time_of_parent][i_dr_of_parent][i_parent_sub];
				TBD_nodes[offset].parents_possible[i_parent] = true;
				i_parent++;
			}
		}
	}

	bool* which_TD;
	which_TD = new bool[nb_nodes];
	for (int i_node = 0; i_node < nb_nodes; i_node++){
		which_TD[i_node] = true;
	}

	// Creating the trajectory list
	int** trajectories_list;
	trajectories_list = new int* [(int)(pow(2, T))];
	for (int i_trajectory = 0; i_trajectory < pow(2, T); i_trajectory++) {
		trajectories_list[i_trajectory] = new int[T];
	}

	for (int i_trajectory = 0; i_trajectory < pow(2, T); i_trajectory++) {
		for (int t = 0; t < T; t++) {
			trajectories_list[i_trajectory][t] = 0;
		}
	}

	double* trajectories_list_values;
	trajectories_list_values = new double[(int)(pow(2, T))];

	int* trajectories_list_indices;
	trajectories_list_indices = new int[(int)(pow(2, T))];
	for (int i_trajectory = 0; i_trajectory < pow(2, T); i_trajectory++) {
		trajectories_list_indices[i_trajectory] = i_trajectory;
		//std::cout << "Traj " << i_trajectory << ": " << trajectories_list_indices[i_trajectory] << "\n";
	}

	// Filling the trajectory list
	for (int t = 0; t < T; t++) {
		for (int i_interval_start = pow(2, T - 1 - t); i_interval_start < pow(2, T); i_interval_start += pow(2, T - t)) {
			for (int i_within_interval = 0; i_within_interval < pow(2, T - 1 - t); i_within_interval++) {
				trajectories_list[i_interval_start + i_within_interval][t] = 1;
			}
		}
	}

	for (int i_trajectory = 0; i_trajectory < pow(2, T); i_trajectory++) {
		for (int t = 0; t < T; t++) {
			//std::cout << trajectories_list[i_trajectory][t] << " ";
		}
		//std::cout << "\n";
	}

	int* how_many_dec_there;
	how_many_dec_there = new int[T];
	for (int t = 0; t < T; t++)
	{
		how_many_dec_there[t] = 0;
	}

	double* start_facilities;
	start_facilities = new double[nb_locations];
	for (int i_l = 0; i_l < nb_locations; i_l++)
	{
		start_facilities[i_l] = 0;
	}

	double* end_facilities;
	end_facilities = new double[nb_locations];
	for (int i_l = 0; i_l < nb_locations; i_l++) {
		end_facilities[i_l] = 0;
	}

	double*** policies_x;
	policies_x = new double** [T];

	for (int t = 0; t < T; t++) {
		policies_x[t] = new double* [max_dec_per_period];

		for (int i_dec = 0; i_dec < max_dec_per_period; i_dec++) {
			policies_x[t][i_dec] = new double[nb_locations];
		}
	}

	int* how_many_unique_dec_there;
	how_many_unique_dec_there = new int[T];

	problem_state* myproblem;
	myproblem = new problem_state;
	myproblem->arcs = arcs;
	myproblem->T = T;
	myproblem->b = b;
	myproblem->m_cost = m_cost;
	myproblem->nb_nodes = nb_nodes;
	myproblem->how_many_dec_there = how_many_dec_there;
	myproblem->how_many_unique_dec_there = how_many_unique_dec_there;
	myproblem->how_many_i_dr = how_many_i_dr;
	myproblem->nb_locations = nb_locations;
	myproblem->nb_customers = nb_customers;
	myproblem->c = c;
	myproblem->nb_of_arcs = nb_of_arcs;
	myproblem->nb_TBD_nodes = nb_TBD_nodes;
	myproblem->arcs = arcs;
	myproblem->nodes = nodes;
	myproblem->TBD_nodes = TBD_nodes;
	myproblem->customers = customers;
	myproblem->factories = factories;
	myproblem->map_TB_to_node = map_TB_to_node;
	myproblem->map_TBD_to_node = map_TBD_to_node;
	myproblem->policies_x = policies_x;
	myproblem->revenue = revenue;
	myproblem->output_file = results_filename;
	myproblem->log_file = log_filename;
	myproblem->policies_file = policies_filename;
	myproblem->uncertainty_level = uncertainty_level;
	myproblem->nb_factors = 4;
	myproblem->Phi = Phi;
	myproblem->demand_factors = demand_factors;
	myproblem->i_instance = i_instance;
	myproblem->env = new GRBEnv();

	int start_period = 0;
	int end_period = T - 1;
	//myfile << "Instance: " << i_instance << "\n";
	myfile.flush();
	double almost_dynamic_val;

	std::chrono::steady_clock::time_point begin_t;// = std::chrono::steady_clock::now();
	std::chrono::steady_clock::time_point end_t;// = std::chrono::steady_clock::now();
	time_t start, end;
	double updating_time;
	double solution_time;

	if (method_used <= 1) {// Dynamic solution for 1h
		myproblem->TimeLimit = 3600;
		//start = clock();
		begin_t = std::chrono::steady_clock::now();
		if (method_used == 0) {
			almost_dynamic_val = dynamic_solution(which_TD, start_period, end_period, start_facilities, end_facilities, myproblem, 1, true);
		}
		else {
			almost_dynamic_val = dynamic_solution_no_IU(which_TD, start_period, end_period, start_facilities, end_facilities, myproblem, 1, true);
		}
		end_t = std::chrono::steady_clock::now();
		solution_time = (double)(std::chrono::duration_cast<std::chrono::milliseconds>(end_t - begin_t).count()) / 1000;
		updating_time = 0;
		
		double** per_period_updating_times;
		per_period_updating_times = new double*[(int)(1)];
		for (int i = 0; i < 1; i++) {
			per_period_updating_times[i] = new double[T];
			for (int j = 0; j < T; j++) {
				per_period_updating_times[i][j] = 0;
			}
		}
		
		begin_t = std::chrono::steady_clock::now();
		backward_updating(myproblem, per_period_updating_times, 0, true);
		end_t = std::chrono::steady_clock::now();
		updating_time = (double)(std::chrono::duration_cast<std::chrono::milliseconds>(end_t - begin_t).count()) / 1000;

		double** unique_policies;
		unique_policies = new double* [(int)(improvement_rounds + 1)];
		for (int i_r = 0; i_r < improvement_rounds + 1; i_r++) {
			unique_policies[i_r] = new double[T];
		}

		int i_round_index = 0;
		count_unique_policies_to_array(myproblem, unique_policies, i_round_index);
		// Check if there is any initial solution found
		if(unique_policies[0][0] > 0){
			myfile << 1 << std::endl;
			myfile.flush(); 
		}else{
			myfile << 0 << std::endl;
			myfile.flush(); 
		}
		
		myfile << TBD_nodes[0].han_plus_cost_to_go << " " << almost_dynamic_val << " " << 0 << " " << solution_time << " " << updating_time << " ";
		for (int t = 0; t < T; t++){
			myfile << per_period_updating_times[0][t] << " ";
		}
		myfile.flush(); 
		myfile << std::endl;

		for (int t = 0; t < T; t++) {
			myfile << unique_policies[0][t] << " ";
			myfile.flush();
		}
		myfile << std::endl;
		myfile.flush();
	} else if (method_used <= 6){// Dynamic with IU 10min, backward recursions with IU, improvement with IU, tracking
		myproblem->TimeLimit = 600; // TODO
		int start_method = 0; // (method_used - 1) / 18;
		int worst_TD_method = method_used - 2; // ((method_used - 1) % 18) / 6;
		int how_many_were_used = 0;
		//int improvement_method  = (((method_used - 1) % 18) % 6) % 6 + 1;

		//std::cout << "start method: " << start_method << std::endl;
		std::cout << "worst_TD_method: " << worst_TD_method << std::endl;
		//std::cout << "improvement_method: " << improvement_method << std::endl;
		std::chrono::steady_clock::time_point time_start_method = std::chrono::steady_clock::now();
		std::chrono::steady_clock::time_point time_end_method;

		double* real_values;
		real_values = new double[(int)(improvement_rounds + 1)];

		double* backward_recursion_values;
		backward_recursion_values = new double[(int)(improvement_rounds + 1)];

		double** updating_times;
		updating_times = new double*[(int)(improvement_rounds + 1)];
		for (int i = 0; i < improvement_rounds + 1; i++) {
			real_values[i] = 0;
			backward_recursion_values[i] = 0;
			updating_times[i] = new double[3];
			for (int j = 0; j < 3; j++) {
				updating_times[i][j] = 0;
			}
		}

		double** per_period_updating_times;
		per_period_updating_times = new double*[(int)(improvement_rounds + 1)];
		for (int i = 0; i < improvement_rounds + 1; i++) {
			per_period_updating_times[i] = new double[T];
			for (int j = 0; j < T; j++) {
				per_period_updating_times[i][j] = 0;
			}
		}

		double** unique_policies;
		unique_policies = new double* [(int)(improvement_rounds + 1)];
		for (int i_r = 0; i_r < improvement_rounds + 1; i_r++) {
			unique_policies[i_r] = new double[T];
		}

		// If you do improvement without IU and there is no change, you switch to with IU
		bool b_rec_IU;
		int changed_start_IU_time = 0;
		bool start_IU = false;
		double previous_value;
		b_rec_IU = true;

		int i_round_index = 0;
		begin_t = std::chrono::steady_clock::now();
		if(start_method < 0.5){
			almost_dynamic_val = dynamic_solution_no_IU(which_TD, start_period, end_period, start_facilities, end_facilities, myproblem, 1, true);
		} else {
			almost_dynamic_val = static_solution_no_IU(which_TD, start_period, end_period, start_facilities, end_facilities, myproblem, 1, true);
		}
		end_t = std::chrono::steady_clock::now();
		solution_time = (double)(std::chrono::duration_cast<std::chrono::milliseconds>(end_t - begin_t).count()) / 1000;
		updating_times[i_round_index][1] = solution_time;

		export_policies(myproblem);
		print_all_policies(myproblem, false);
	}

	myfile.close();
	return 0;
}
