using DelimitedFiles
using Statistics
using Printf
using Plots
using LinearAlgebra

N_instances = 50
N_methods = 7
improvement_rounds = 100
T = 10
for how_many_worst = 2:10
    feasibility = zeros(Bool, N_instances, N_methods)
    objective_values = zeros(Float64, N_instances, N_methods, improvement_rounds + 1, 2)
    run_times = zeros(Float64, N_instances, N_methods, improvement_rounds + 1, T + 3)
    number_of_policies = zeros(Int64, N_instances, N_methods, improvement_rounds + 1, T)
    f_local = open(string("results_all.txt"), "w")

    max_time = 12*3600
    time_stamp_every = 10
    time_scale = collect(0:time_stamp_every:max_time)
    time_objective = zeros(Float64, N_instances, N_methods, length(time_scale))
    time_rounds = zeros(Float64, N_instances, N_methods, length(time_scale))
    time_objective_medians = zeros(Float64, N_methods, length(time_scale))

    for i_attempt = 1:N_instances
        #=
        for i_method = 0:2
            f = open(string("results_", i_attempt, "_", i_method, "_", 0, ".txt"))
            lines = readlines(f)
            offset_line = 2
            line = [parse(Float64, ss) for ss in split(lines[offset_line])]
            objective_values[i_attempt, i_method + 1, 1, :] = line[1:2]
            time_objective[i_attempt, i_method + 1, :] .= objective_values[i_attempt, i_method + 1, 1, 1]
            close(f)
        end=#

        for i_method = 0:6
            results_filename = ""
            if (i_method < 2)
                results_filename = string("results_", i_attempt, "_", i_method, "_", 0, ".txt")
            else
                results_filename = string("results_", i_attempt, "_", i_method, "_", how_many_worst, ".txt")
            end

            if (isfile(results_filename))
                f = open(results_filename)
                lines = readlines(f)
                
                if (length(lines) < 1)
                    print(i_attempt, " ", i_method, "file empty\n")
                else
                    solved_or_not = parse(Float64, lines[1])
                    #print(solved_or_not)

                    if (solved_or_not > 0.5)
                        if (i_method < 2)
                            feasibility[i_attempt, i_method + 1] = true
                            time_vs_obj = readdlm(string("incumbent_", i_attempt, "_", i_method, ".txt"))
                            time_vs_obj[:, 1] = time_vs_obj[:, 1]

                            for i_timestamp = 1:length(time_scale)
                                leftover_stuff = time_vs_obj[time_vs_obj[:, 1] .< time_scale[i_timestamp] + 0.1, 2]
                                
                                if (length(leftover_stuff) > 0)
                                    time_objective[i_attempt, i_method + 1, i_timestamp] = leftover_stuff[end]
                                    time_rounds[i_attempt, i_method + 1, i_timestamp] = size(leftover_stuff, 1)
                                end
                            end
                        else
                            feasibility[i_attempt, i_method + 1] = true
                            time_vs_obj = readdlm(string("time_obj_", i_attempt, "_", i_method, "_", how_many_worst, ".txt"))
                            time_vs_obj[:, 1] = time_vs_obj[:, 1] .- time_vs_obj[1, 1]

                            for i_timestamp = 1:length(time_scale)
                                leftover_stuff = time_vs_obj[time_vs_obj[:, 1] .< time_scale[i_timestamp] + 0.1, 2]
                                
                                if (length(leftover_stuff) > 0)
                                    time_objective[i_attempt, i_method + 1, i_timestamp] = leftover_stuff[end]
                                    time_rounds[i_attempt, i_method + 1, i_timestamp] = size(leftover_stuff, 1)
                                end
                            end
                        end
                    end

                    close(f)
                end
            else
                print(i_attempt, " ", i_method, " nofile \n")
            end
        end
    end

    # correct the results for the improvement methods
    start_improvement = 61

    for i_method = 3:7
        for i_attempt = 1:N_instances
            time_objective[i_attempt, i_method, start_improvement:end] = time_objective[i_attempt, i_method, 1:(length(time_scale) - start_improvement + 1)]
        end
    end

    #=
    # what are the instances where all the methods are feasible
    till_when_zero = zeros(Int64, N_instances, N_methods)

    for i_attempt = 1:N_instances
        for i_method = 4:13
            till_when_what = cumprod(abs.(time_objective[i_attempt, i_method, :]) .<= 0.001)
            till_when_zero[i_attempt, i_method] = sum(till_when_what)
        end
    end
    =#

    #print(till_when_zero)
    for i_method = 1:7
        for i_timestamp = 1:length(time_scale)
            time_objective_medians[i_method, i_timestamp] = Statistics.median(time_objective[:, i_method, i_timestamp], dims=1)[1]
        end
    end

    # plotting
    time_objective_normalized = zeros(Float64, N_instances, N_methods, length(time_scale))
    for i_attempt = 1:N_instances
        minval = Statistics.minimum(time_objective[i_attempt, :, end])
        #print(minval,"\n")

        for i_timestamp = 1:length(time_scale)
            time_objective_normalized[i_attempt, :, i_timestamp] = (time_objective[i_attempt, :, i_timestamp] .- minval) ./ abs.(minval) * 100
        end
    end

    print("\n\n")
    for i_method = 2:7
        println(time_objective_normalized[:, i_method, end])
    end

    time_objective_normalized_medians = zeros(Float64, N_methods, length(time_scale))
    for i_method = 1:7
        for i_timestamp = 1:length(time_scale)
            time_objective_normalized_medians[i_method, i_timestamp] = Statistics.mean(time_objective_normalized[:, i_method, i_timestamp], dims=1)[1]
        end
    end

    labels = ["1 worst" "min(k, worst)" "worst" "max(k, worst)" "k"]
    colornames = ["Blue", "Red", "Green", "Orange", "Violet"]

    plot()
    where_start = 10
    time_scale = time_scale ./ 3600
    #plot!(time_scale, time_objective_normalized_medians[1, :], seriestype = :line, label = "CCG dynamic 1h", linestyle = :dot, linecolor = "Black")
    plot!(time_scale[where_start:end], time_objective_normalized_medians[1, where_start:end], seriestype = :line, label = "IU dynamic 1h", linestyle = :dot, linecolor = "Black")
    plot!(time_scale[where_start:end], time_objective_normalized_medians[2, where_start:end], seriestype = :line, label = "No IU dynamic 1h", linestyle = :dot, linecolor = "Brown")

    for i_method = 1:5
        column_index = 2 + i_method
        this_label = ""

        this_label = string(labels[i_method], " CCG")
        plot!(time_scale[start_improvement:end], time_objective_normalized_medians[column_index, start_improvement:end], seriestype = :line, label = this_label, linecolor = colornames[i_method])
    end

    plot!(legend = :outertopright)
    xlabel!("Time (h)")
    ylabel!("Mean relative obj val")
    ylims!((0, 150))

    savefig(string("obj_vs_time_k=", how_many_worst, ".pdf"))
end
#Statistics.median(objective_values[:, 1:2, 1, 1], dims=1)
#print(time_objective[end, 4:13, end])
#print(time_objective_normalized[end, 4:13, end])

# plotting the number of rounds per time
#=
mean_rounds = zeros(Float64, N_methods, length(time_scale))
for i_method = 2:6
    for i_timestamp = 1:length(time_scale)
        mean_rounds[i_method + 1, i_timestamp] = Statistics.mean(time_rounds[:, i_method + 1, i_timestamp], dims=1)[1]
    end
end

plot(time_scale, transpose(mean_rounds), seriestype = :line, label = ["1 worst" "min(k, worst)" "worst" "max(k, worst)" "k"])
savefig("rounds_vs_time.pdf")=#