using DelimitedFiles
using Statistics
using Printf
using Plots
using LinearAlgebra

N_instances = 50
N_methods = 7
improvement_rounds = 100
T = 10
k_variants = 9

feasibility = zeros(Bool, N_instances, N_methods)
objective_values = zeros(Float64, N_instances, N_methods, improvement_rounds + 1, k_variants)
#f_local = open(string("results_all.txt"), "w")

max_time = 12*3600
time_stamp_every = 10
time_scale = collect(0:time_stamp_every:max_time)
time_objective = zeros(Float64, N_instances, N_methods, length(time_scale), k_variants)
time_objective_normalized = zeros(Float64, N_instances, N_methods, length(time_scale), k_variants)
time_objective_normalized_medians = zeros(Float64, N_methods, length(time_scale), k_variants)
#time_objective_medians = zeros(Float64, N_methods, length(time_scale), k_variants)

for how_many_worst = 2:10
    for i_attempt = 1:N_instances
        for i_method = 0:6
            results_filename = ""
            if (i_method < 2)
                results_filename = string("results_", i_attempt, "_", i_method, "_", 0, ".txt")
            else
                results_filename = string("results_", i_attempt, "_", i_method, "_", how_many_worst, ".txt")
            end

            if (isfile(results_filename))
                f = open(results_filename)
                lines = readlines(f)
                
                if (length(lines) < 1)
                    print(i_attempt, " ", i_method, "file empty\n")
                else
                    solved_or_not = parse(Float64, lines[1])
                    #print(solved_or_not)

                    if (solved_or_not > 0.5)
                        if (i_method < 2)
                            feasibility[i_attempt, i_method + 1] = true
                            time_vs_obj = readdlm(string("incumbent_", i_attempt, "_", i_method, ".txt"))
                            time_vs_obj[:, 1] = time_vs_obj[:, 1]

                            for i_timestamp = 1:length(time_scale)
                                leftover_stuff = time_vs_obj[time_vs_obj[:, 1] .< time_scale[i_timestamp] + 0.1, 2]
                                
                                if (length(leftover_stuff) > 0)
                                    time_objective[i_attempt, i_method + 1, i_timestamp, 1] = leftover_stuff[end]
                                    #time_rounds[i_attempt, i_method + 1, i_timestamp, 1] = size(leftover_stuff, 1)
                                end
                            end
                        else
                            feasibility[i_attempt, i_method + 1] = true
                            time_vs_obj = readdlm(string("time_obj_", i_attempt, "_", i_method, "_", how_many_worst, ".txt"))
                            time_vs_obj[:, 1] = time_vs_obj[:, 1] .- time_vs_obj[1, 1]

                            for i_timestamp = 1:length(time_scale)
                                leftover_stuff = time_vs_obj[time_vs_obj[:, 1] .< time_scale[i_timestamp] + 0.1, 2]
                                
                                if (length(leftover_stuff) > 0)
                                    time_objective[i_attempt, i_method + 1, i_timestamp, how_many_worst - 1] = leftover_stuff[end]
                                    #time_rounds[i_attempt, i_method + 1, i_timestamp, how_many_worst - 1] = size(leftover_stuff, 1)
                                end
                            end
                        end
                    end

                    close(f)
                end
            else
                print(i_attempt, " ", i_method, " nofile \n")
            end
        end
    end
end

# copying the results for the first two methods
for i_method = 1:2
    for i_copy = 2:k_variants
        time_objective[:, i_method, :, i_copy] .= time_objective[:, i_method, :, 1]
    end
end

# correct the results for the improvement methods
start_improvement = 61
time_objective[:, 3:7, start_improvement:end, :] .= time_objective[:, 3:7, 1:(length(time_scale) - start_improvement + 1), :]

for how_many_worst = 2:10
    for i_attempt = 1:N_instances
        minval = Statistics.minimum(time_objective[i_attempt, :, end, :])
        #println(minval)
        time_objective_normalized[i_attempt, :, :, how_many_worst - 1] = (time_objective[i_attempt, :, :, how_many_worst - 1] .- minval) ./ abs.(minval) * 100
    end

    for i_method = 1:7
        for i_timestamp = 1:length(time_scale)
            time_objective_normalized_medians[i_method, i_timestamp, how_many_worst - 1] = Statistics.mean(time_objective_normalized[:, i_method, i_timestamp, how_many_worst - 1])[1]
        end
    end

    #println(time_objective_normalized_medians[:, start_improvement, how_many_worst - 1])
end

where_start = 10
labels = ["one_worst" "min(k,worst)" "all_worst" "max(k,worst)" "k"]
colornames = ["Blue", "Red", "Green", "Orange", "Violet"]
time_scale = time_scale ./ 3600

time_objective_normalized_medians

for i_method = 1:5
    column_index = 2 + i_method

    plot()
    where_start = 10
    plottitle = string("obj_vs_time:", labels[i_method])
    plot!(time_scale[where_start:end], time_objective_normalized_medians[1, where_start:end, 2], title = plottitle, seriestype = :line, label = "CCG dynamic", linestyle = :dot, linecolor = "Black")
    plot!(time_scale[where_start:end], time_objective_normalized_medians[2, where_start:end, 2], title = plottitle, seriestype = :line, label = "No IU dynamic", linestyle = :dot, linecolor = "Grey")

    for how_many_worst = 2:10
        this_label = string("k = ",  how_many_worst)
        println(time_objective_normalized_medians[column_index, start_improvement, how_many_worst - 1])
        plot!(time_scale[start_improvement:end], time_objective_normalized_medians[column_index, start_improvement:end, how_many_worst - 1], seriestype = :line, label = this_label, linecolor = "Red", lineopacity = (how_many_worst / 10)) #colornames[how_many_worst - 1])
    end

    plot!(legend = :outertopright)
    xlabel!("Time (h)")
    ylabel!("Mean relative obj val")
    ylims!((-50, 150))
    savefig(string("obj_vs_time-", labels[i_method], ".pdf"))
end