﻿// ConsoleApplication1.cpp : Ten plik zawiera funkcję „main”. W nim rozpoczyna się i kończy wykonywanie programu.

#include <stdlib.h>
#include <string>
#include <math.h>
#include <fstream>
#include <iostream>
#include <ctime>
#include <sstream>
#include "gurobi_c++.h"
#include <sstream>
#include <chrono>

using namespace std;

bool NotConnected(int number_of_nodes, bool** A_square_included, int s, int t) { // function computing distance from s to t using Dijkstra's algorithm, it doesn't require arc lengths because attainability is all that matters
	double* dist_from_s;
	dist_from_s = new double[number_of_nodes]; // array storing the computed shortest distances from node s
	bool* visited;
	visited = new bool[number_of_nodes]; // array storing information if a given node already visited by Dijkstra's algorithm

	//std::cout << "Path: " << s << " " << t << "\n";
	for (int i = 0; i < number_of_nodes; i++){
		dist_from_s[i] = 1000; // infinity distance from s initially
		visited[i] = false; // nodes unvisited
	}

	dist_from_s[s] = 0; // distance is zero for node s
	visited[s] = true; // node s has already been visited

	int current_node = s;
	bool attainable;
	int chosen_next;
	double dist_chosen_next;

	for (int i = 0; i < number_of_nodes; i++){
		dist_chosen_next = 100; // the distance to the next node to visit is some large number 
		chosen_next = current_node; // for a start we take that the next node to visit is s itself

		for (int j = 0; j < number_of_nodes; j++){
			if ((!visited[j]) && (dist_from_s[current_node] + 1 < dist_from_s[j]) && ((A_square_included[current_node][j]) || (A_square_included[j][current_node]))){
				// a candidate for a new node to visit should be previously unvisited, the distance should be better than the one in the table so far, and there should be an undirected arc leading to it in the current graph
				dist_from_s[j] = dist_from_s[current_node] + 1; // update the shortest path to the new node
			}
		}

		for (int k = 0; k < number_of_nodes; k++){
			if ((!visited[k]) && (dist_from_s[k] < dist_chosen_next)){
				dist_chosen_next = dist_from_s[k]; // to look for the closest unvisited node
				chosen_next = k; // next node to visit updated
			}
		}

		if (chosen_next == t) {// it is possible to reach t going from s
			attainable = true; // t attainable from s
			break; // stop the loop
		}
		else if (chosen_next == current_node){	
			attainable = false; // there is no new unvisited node that can be reached from s
			break;
		} else { 
			current_node = chosen_next;
			visited[current_node] = true; // the new node to visit will be updated that it shouldn't be visited anymore
		}
	}

	return (!attainable);
}

int main(int argc, char* argv[]) {
	std::vector<std::string> args(argc);
	for (int i = 0; i < argc; ++i) {
		args[i] = argv[i];
	};

	int number_of_nodes = atoi(argv[1]);
	int i_instance = atoi(argv[2]);
	int output_flag = 1;
	int neighbors_left_threshold = 14;
	srand(i_instance);

	int T = number_of_nodes / 10 * 4; // max number of time periods to traverse the grapho
	int max_Gamma = T; // max budgjet
	int N_sim = 1000; // Number of simulated passes through the graph

	ofstream outfile;
	ostringstream o_filename;
	o_filename << "routing_output_" << number_of_nodes << "_" << i_instance << ".txt";
	outfile.open(o_filename.str().c_str());

	// All the code here is to construct the graph and even though looks ugly, it's fast
	//define the distances matrix
	double** V;
	V = new double* [number_of_nodes];

	ofstream outfile_nodes; // Create a file to output the node data (for plotting)
	ostringstream o_filename_nodes;
	o_filename_nodes << "nodes_data_" << number_of_nodes << "_" << i_instance << ".txt";
	outfile_nodes.open(o_filename_nodes.str().c_str());
	double cord_a, cord_b; // coordinate stuff
	bool node_ok; // acceptance

	for (int j = 0; j < number_of_nodes; j++) {
		V[j] = new double[2]; 

		if ((j > 0) && (j < number_of_nodes - 1)){ // Insert here only nodes that do fall into the tube
			node_ok = false;

			do {
				cord_a = ((double)rand() / (RAND_MAX)) * 10; // cordinate sampling
				cord_b = ((double)rand() / (RAND_MAX)) * 10;

				if (!(((cord_a > 2) && (cord_a < 4) && (cord_b < 8)) || ((cord_a > 6) && (cord_a < 8) && (cord_b > 2)))){ // selection
					node_ok = true;
				}

			} while (!node_ok); //std::cout << cord_a << " " << cord_b << "\n";

			V[j][0] = cord_a;
			V[j][1] = cord_b;
		} else if (j == 0){
			V[j][0] = 0;
			V[j][1] = 0;
		} else {
			V[j][0] = 10;
			V[j][1] = 10;
		}

		outfile_nodes << (j + 1) << " " << V[j][0] << " " << V[j][1] << "\n";
	}

	outfile_nodes.close();
	return 0;
}