using DelimitedFiles
using Statistics
using Printf
using Plots
using LinearAlgebra

# extracting obj values
n = 100
max_Gamma = 90
objective_values = zeros(Float64, max_Gamma + 1, 50, 3)
instance_used = 0
global i_ms = 1

for i_instance = 1:50
    i_method = 0

    for i_B = 0:1:max_Gamma
        fname = string("routing_output_", n, "_", i_instance ,"_" , i_method, "_", i_B,  ".txt")
        
        if isfile(fname)
            f = open(fname)
            lines = readlines(f)
            first_line = [parse(Float64, ss) for ss in split(lines[1])]
            objective_values[i_B + 1, i_instance, 1] = first_line[1]
            objective_values[i_B + 1, i_instance, 2] = first_line[3]
            close(f)
        else
            print("Missing file: ", fname, "\n")
            f_missing_name = string("missing_command", i_ms,  ".txt")
            f_missing = open(f_missing_name, "w")
            write(f_missing, string(i_instance, " ", 0 , " ", i_B))
            close(f_missing)

            f_missing_name = string("missing_file", i_ms,  ".txt")
            f_missing = open(f_missing_name, "w")
            write(f_missing, string("nodes_data_100_", i_instance, ".txt"))
            close(f_missing)

            global i_ms += 1
        end
    end
end

i_method = 1
working_instances = zeros(Bool, 50)

for i_instance = 1:50
    print("Instance: ", i_instance, "\n")
    f = open(string("routing_output_", n, "_", i_instance ,"_" , i_method, ".txt"))
    lines = readlines(f)
    #print(length(lines),"\n")

    if (length(lines) > 41)
        working_instances[i_instance] = true

        for i_B = 0:1:max_Gamma
            current_line = [parse(Float64, ss) for ss in split(lines[i_B + 1])]
            objective_values[i_B + 1, i_instance, 3] = current_line[1]
        end
    else
        working_instances[i_instance] = false
        print("Fucked instance: ", i_instance, "\n")
    end
    
    close(f)
end

mean_lengths = Statistics.mean(objective_values[:, working_instances, :], dims=2)
gamma_scale = collect(0:1:max_Gamma)
plot(gamma_scale, reshape(mean_lengths, (size(gamma_scale, 1), 3)), seriestype = :line, label = ["static" "crystal ball" "adaptive"])
savefig(string("routing_n=", n, ".pdf"))

#=

n_range = collect(10:10:110)
p_range = collect(0.05:0.05:0.3)
MAX_Gamma = 2 * Int64(n_range[length(n_range)] / 10) * 4
leftover_gammas = zeros(Float64, length(n_range), 50, MAX_Gamma + 1, length(p_range))
leftover_gammas_file = zeros(Bool, length(n_range), 50, MAX_Gamma + 1, length(p_range))

for i_n = 1:length(n_range)# 30:10:110
    n = n_range[i_n]
    T = Int64(n / 10) * 4
    max_Gamma = 2 * T
    objective_values = zeros(Float64, max_Gamma + 1, 50, 3)

    for i_instance = 1:50
        for i_B = 0:1:max_Gamma
            for i_p = 1:length(p_range)-1 #0.05:0.05:0.3
                p = p_range[i_p]
                f = open(string("arc_statistics_", n, "_", i_instance, "_", i_B + 1, "_", p, ".txt"))
                lines = readlines(f)
                splitlines = split(lines[1])
                leftover_gammas[i_n, i_instance, i_B + 1, i_p] = parse(Float64, splitlines[1])
            end

            close(f)
        end
    end

    #print(objective_values)
end =#

# now making the heatmap plot

function routing_heatmap(i_instance::Int64, i_B::Int64)
    #downloading the nodes coordinates
    nodes_coordinates = zeros(Float64, n, 2)
    f = open(string("nodes_data_", n, "_", i_instance, ".txt"))
    lines = readlines(f)
    for offset_line = 1:length(lines)
        line = [parse(Float64, ss) for ss in split(lines[offset_line])]
        nodes_coordinates[offset_line, :] = line[2:3]
    end
    close(f)

    # now downloading the arc statistics
    f = open(string("arc_statistics_", n, "_", i_instance, "_", i_B, ".txt"))
    lines = readlines(f)
    arcs_indices_frequency = zeros(Int64, length(lines) - 1, 3)
    first_line = [parse(Float64, ss) for ss in split(lines[1])]
    left_gamma = first_line[1]
    p = first_line[2]

    for offset_line = 2:length(lines)
        line = [parse(Int64, ss) for ss in split(lines[offset_line])]
        arcs_indices_frequency[offset_line - 1, :] = line
        arcs_indices_frequency[offset_line - 1, 1:2] = arcs_indices_frequency[offset_line - 1, 1:2] .+ 1
    end
    close(f)

    # now downloading the static thingy
    f = open(string("static_", n, "_", i_instance, "_", i_B, ".txt"))
    lines = readlines(f)
    static_path = zeros(Int64, length(lines), 2)
    for offset_line = 1:length(lines)
        line = [parse(Int64, ss) for ss in split(lines[offset_line])]
        static_path[offset_line, :] = line .+ 1
    end
    close(f)

    # now plotting the heatmap
    plt = plot(xlims = (0, 10), ylims = (0, 10), title = string("Instance: ", i_instance, ",initial B: ", i_B))#, ", left B: ", left_gamma, ", probability", p))

    # plot all arcs in the plot first
    for i_arc = 1:size(arcs_indices_frequency, 1)
        if (!(arcs_indices_frequency[i_arc, 1] == arcs_indices_frequency[i_arc, 2]))
            coords_x = [nodes_coordinates[arcs_indices_frequency[i_arc, 1], 1], nodes_coordinates[arcs_indices_frequency[i_arc, 2], 1]]   
            coords_y = [nodes_coordinates[arcs_indices_frequency[i_arc, 1], 2], nodes_coordinates[arcs_indices_frequency[i_arc, 2], 2]]
            plot!(plt, coords_x, coords_y, legend = false, linewidth = 0.25, color = "gray")
        end
    end

    # plot the static path
    for i_arc = 1:size(static_path, 1)
        coords_x = [nodes_coordinates[static_path[i_arc, 1], 1], nodes_coordinates[static_path[i_arc, 2], 1]]   
        coords_y = [nodes_coordinates[static_path[i_arc, 1], 2], nodes_coordinates[static_path[i_arc, 2], 2]]
        plot!(plt, coords_x, coords_y, legend = false, linewidth = 6, color = "blue")
    end

    # plot the adaptive paths
    for i_arc = 1:size(arcs_indices_frequency, 1)
        if (!(arcs_indices_frequency[i_arc, 1] == arcs_indices_frequency[i_arc, 2]))
            coords_x = [nodes_coordinates[arcs_indices_frequency[i_arc, 1], 1], nodes_coordinates[arcs_indices_frequency[i_arc, 2], 1]]   
            coords_y = [nodes_coordinates[arcs_indices_frequency[i_arc, 1], 2], nodes_coordinates[arcs_indices_frequency[i_arc, 2], 2]]
            #print(coords_x)
            #print(coords_y)
            if (arcs_indices_frequency[i_arc, 3] > 0)
                thickness = Int64(ceil(Float64(arcs_indices_frequency[i_arc, 3]) / 1000 * 5))
            else
                thickness = 0
            end

            #print(thickness, "\n")
            if (thickness > 0)
                plot!(plt, coords_x, coords_y, legend = false, linewidth = thickness, color = "red")
            end
        end
    end
    plt
    savefig(string("arc_statistics_", n, "_", i_instance, "_", i_B, ".pdf"))
end


# making the GIF file
#=
for i_instance = 1:10
    anim = @animate for i_B in 0:40
        routing_heatmap(i_instance, i_B)
    end
    gif(anim, string("anim_instance_", i_instance, "_fps2.gif"), fps = 4)
end=#


i_instance = 9
for i_B in 0:40
    routing_heatmap(i_instance, i_B)
end
