#=
using DelimitedFiles
using Statistics
using Printf
using Plots
using LinearAlgebra

# extracting obj values
n = 100
max_Gamma = 90
objective_values = zeros(Float64, max_Gamma + 1, 50, 3)
instance_used = 0
global i_ms = 1

for i_instance = 1:50
    i_method = 0

    for i_B = 0:1:max_Gamma
        fname = string("routing_output_", n, "_", i_instance ,"_" , i_method, "_", i_B,  ".txt")
        
        if isfile(fname)
            f = open(fname)
            lines = readlines(f)
            first_line = [parse(Float64, ss) for ss in split(lines[1])]
            objective_values[i_B + 1, i_instance, 1] = first_line[1]
            objective_values[i_B + 1, i_instance, 2] = first_line[3]
            close(f)
        else
            print("Missing file: ", fname, "\n")
            f_missing_name = string("missing_command", i_ms,  ".txt")
            f_missing = open(f_missing_name, "w")
            write(f_missing, string(i_instance, " ", 0 , " ", i_B))
            close(f_missing)

            f_missing_name = string("missing_file", i_ms,  ".txt")
            f_missing = open(f_missing_name, "w")
            write(f_missing, string("nodes_data_100_", i_instance, ".txt"))
            close(f_missing)

            global i_ms += 1
        end
    end
end

i_method = 1
working_instances = zeros(Bool, 50)

for i_instance = 1:50
    print("Instance: ", i_instance, "\n")
    f = open(string("routing_output_", n, "_", i_instance ,"_" , i_method, ".txt"))
    lines = readlines(f)
    #print(length(lines),"\n")

    if (length(lines) > 41)
        working_instances[i_instance] = true

        for i_B = 0:1:max_Gamma
            current_line = [parse(Float64, ss) for ss in split(lines[i_B + 1])]
            objective_values[i_B + 1, i_instance, 3] = current_line[1]
        end
    else
        working_instances[i_instance] = false
        print("Fucked instance: ", i_instance, "\n")
    end
    
    close(f)
end

mean_lengths = Statistics.mean(objective_values[:, working_instances, :], dims=2)
gamma_scale = collect(0:1:max_Gamma)
plot(gamma_scale, reshape(mean_lengths, (size(gamma_scale, 1), 3)), xlabel = "Uncertainty budget", ylabel = "Length", linestyle = [ :solid :dash :dot], linecolor = [:black :black :black], seriestype = :line, label = ["Static" "Crystal ball" "Adaptive"], legend=:bottomright)
savefig(string("routing_n=", n, ".pdf"))


for i_instance = 1:50
    for i_method = 1:3
        if working_instances[i_instance]
            for i_B = 0:1:max_Gamma
                if i_B < max_Gamma && (objective_values[i_B + 1, i_instance, i_method] - 1 > objective_values[i_B + 2, i_instance, i_method])
                    print("Wrong stuff instance: ", i_instance, " method: ", i_method, "\n")
                end
            end
        end
    end
end

=#
# Now the simulation graphs

function routing_heatmap(i_instance::Int64, i_B::Int64)
    #downloading the nodes coordinates
    nodes_coordinates = zeros(Float64, n, 2)
    f = open(string("nodes_data_", n, "_", i_instance, ".txt"))
    lines = readlines(f)
    for offset_line = 1:length(lines)
        line = [parse(Float64, ss) for ss in split(lines[offset_line])]
        nodes_coordinates[offset_line, :] = line[2:3]
    end
    close(f)

    # now downloading the arc statistics
    f = open(string("arc_statistics_", n, "_", i_instance, "_", i_B, ".txt"))
    lines = readlines(f)
    arcs_indices_frequency = zeros(Int64, length(lines) - 1, 3)
    first_line = [parse(Float64, ss) for ss in split(lines[1])]
    left_gamma = first_line[1]
    p = first_line[2]

    for offset_line = 2:length(lines)
        line = [parse(Int64, ss) for ss in split(lines[offset_line])]
        arcs_indices_frequency[offset_line - 1, :] = line
        arcs_indices_frequency[offset_line - 1, 1:2] = arcs_indices_frequency[offset_line - 1, 1:2] .+ 1
    end
    close(f)

    # now downloading the static thingy
    f = open(string("static_", n, "_", i_instance, "_", i_B, ".txt"))
    lines = readlines(f)
    static_path = zeros(Int64, length(lines), 2)
    for offset_line = 1:length(lines)
        line = [parse(Int64, ss) for ss in split(lines[offset_line])]
        static_path[offset_line, :] = line .+ 1
    end
    close(f)

    # now plotting the heatmap
    plt = plot(xlims = (0, 10), ylims = (0, 10))#, title = string("Instance: ", i_instance, ",initial B: ", i_B))#, ", left B: ", left_gamma, ", probability", p))

    # plot all arcs in the plot first
    for i_arc = 1:size(arcs_indices_frequency, 1)
        if (!(arcs_indices_frequency[i_arc, 1] == arcs_indices_frequency[i_arc, 2]))
            coords_x = [nodes_coordinates[arcs_indices_frequency[i_arc, 1], 1], nodes_coordinates[arcs_indices_frequency[i_arc, 2], 1]]   
            coords_y = [nodes_coordinates[arcs_indices_frequency[i_arc, 1], 2], nodes_coordinates[arcs_indices_frequency[i_arc, 2], 2]]
            plot!(plt, coords_x, coords_y, legend = false, linewidth = 0.25, color = "gray", linealpha = 0.3, aspect_ratio = 1.)
        end
    end

    # plot the static path
    for i_arc = 1:size(static_path, 1)
        coords_x = [nodes_coordinates[static_path[i_arc, 1], 1], nodes_coordinates[static_path[i_arc, 2], 1]]   
        coords_y = [nodes_coordinates[static_path[i_arc, 1], 2], nodes_coordinates[static_path[i_arc, 2], 2]]
        plot!(plt, coords_x, coords_y, legend = false, linewidth = 6, color = "blue")
    end

    # plot the adaptive paths
    for i_arc = 1:size(arcs_indices_frequency, 1)
        if (!(arcs_indices_frequency[i_arc, 1] == arcs_indices_frequency[i_arc, 2]))
            coords_x = [nodes_coordinates[arcs_indices_frequency[i_arc, 1], 1], nodes_coordinates[arcs_indices_frequency[i_arc, 2], 1]]   
            coords_y = [nodes_coordinates[arcs_indices_frequency[i_arc, 1], 2], nodes_coordinates[arcs_indices_frequency[i_arc, 2], 2]]
            #print(coords_x)
            #print(coords_y)
            if (arcs_indices_frequency[i_arc, 3] > 0)
                thickness = Int64(ceil(Float64(arcs_indices_frequency[i_arc, 3]) / 1000 * 5))
            else
                thickness = 0
            end

            #print(thickness, "\n")
            if (thickness > 0)
                plot!(plt, coords_x, coords_y, legend = false, linewidth = thickness, color = "red")
            end
        end
    end
    plt
    savefig(string("arc_statistics_", n, "_", i_instance, "_", i_B, ".pdf"))
end

i_instance = 9
for i_B in [0, 15, 30, 45]
    routing_heatmap(i_instance, i_B)
end
