// ConsoleApplication1.cpp : Ten plik zawiera funkcję „main”. W nim rozpoczyna się i kończy wykonywanie programu.

#include <stdlib.h>
#include <string>
#include <math.h>
#include <fstream>
#include <iostream>
#include <ctime>
#include <sstream>
#include "gurobi_c++.h"
#include <sstream>
#include <chrono>

using namespace std;

bool NotConnected(int number_of_nodes, bool** A_square_included, int s, int t) { // function computing distance from s to t using Dijkstra's algorithm, it doesn't require arc lengths because attainability is all that matters
	double* dist_from_s;
	dist_from_s = new double[number_of_nodes]; // array storing the computed shortest distances from node s
	bool* visited;
	visited = new bool[number_of_nodes]; // array storing information if a given node already visited by Dijkstra's algorithm

	//std::cout << "Path: " << s << " " << t << "\n";
	for (int i = 0; i < number_of_nodes; i++){
		dist_from_s[i] = 1000; // infinity distance from s initially
		visited[i] = false; // nodes unvisited
	}

	dist_from_s[s] = 0; // distance is zero for node s
	visited[s] = true; // node s has already been visited

	int current_node = s;
	bool attainable;
	int chosen_next;
	double dist_chosen_next;

	for (int i = 0; i < number_of_nodes; i++){
		dist_chosen_next = 100; // the distance to the next node to visit is some large number 
		chosen_next = current_node; // for a start we take that the next node to visit is s itself

		for (int j = 0; j < number_of_nodes; j++){
			if ((!visited[j]) && (dist_from_s[current_node] + 1 < dist_from_s[j]) && ((A_square_included[current_node][j]) || (A_square_included[j][current_node]))){
				// a candidate for a new node to visit should be previously unvisited, the distance should be better than the one in the table so far, and there should be an undirected arc leading to it in the current graph
				dist_from_s[j] = dist_from_s[current_node] + 1; // update the shortest path to the new node
			}
		}

		for (int k = 0; k < number_of_nodes; k++){
			if ((!visited[k]) && (dist_from_s[k] < dist_chosen_next)){
				dist_chosen_next = dist_from_s[k]; // to look for the closest unvisited node
				chosen_next = k; // next node to visit updated
			}
		}

		if (chosen_next == t) {// it is possible to reach t going from s
			attainable = true; // t attainable from s
			break; // stop the loop
		}
		else if (chosen_next == current_node){	
			attainable = false; // there is no new unvisited node that can be reached from s
			break;
		} else { 
			current_node = chosen_next;
			visited[current_node] = true; // the new node to visit will be updated that it shouldn't be visited anymore
		}
	}

	return (!attainable);
}

int main(int argc, char* argv[]) {
	std::vector<std::string> args(argc);
	for (int i = 0; i < argc; ++i) {
		args[i] = argv[i];
	};

	int number_of_nodes = atoi(argv[1]);
	int i_instance = atoi(argv[2]);
    int i_method = atoi(argv[3]);
    int Gamma = atoi(argv[4]);

	int output_flag = 1;
	int neighbors_left_threshold = 14;
	srand(i_instance);

	int T = number_of_nodes / 10 * 9; // max number of time periods to traverse the grapho
	int max_Gamma = T; // max budgjet
	int N_sim = 1000; // Number of simulated passes through the graph

	// All the code here is to construct the graph and even though looks ugly, it's fast
	//define the distances matrix
	double** V;
	V = new double* [number_of_nodes];
	ostringstream o_filename_nodes; // Read a file with nodes data
	o_filename_nodes << "nodes_data_" << number_of_nodes << "_" << i_instance << ".txt";

    std::fstream nodes_my_file;
	nodes_my_file.open(o_filename_nodes.str().c_str());
    double rubbish_variable;
    for (int j = 0; j < number_of_nodes; j++) {
        V[j] = new double[2]; 
        nodes_my_file >> rubbish_variable;
        nodes_my_file >> V[j][0];
        nodes_my_file >> V[j][1];
    }
    nodes_my_file.close();

	double* A_lengths; // matrix with arc lengths
	A_lengths = new double[(int)(number_of_nodes * number_of_nodes)];

	int** A_ij; // matrix for storying which arc nb is which arc in terms of (i, j) coords
	A_ij = new int* [(int)(number_of_nodes * number_of_nodes)];
	for (int i = 0; i < number_of_nodes * number_of_nodes; i++) {
		A_ij[i] = new int[2];
	}

	double** A_square_distances; // Now computing the distances and listing them all
	A_square_distances = new double* [number_of_nodes];
	for (int i = 0; i < number_of_nodes; i++){
		A_square_distances[i] = new double[number_of_nodes];
	}

	bool* forbidden_arc; // matrix to store the list of forbidden arcs
	forbidden_arc = new bool[(int)(number_of_nodes * number_of_nodes)];

	int i_arc = 0;
	double* arc_middle;
	arc_middle = new double[2];

	for (int i = 0; i < number_of_nodes; i++){
		for (int j = 0; j < number_of_nodes; j++){
			A_ij[i_arc][0] = i;
			A_ij[i_arc][1] = j;
			A_lengths[i_arc] = sqrt(pow(V[i][0] - V[j][0], 2) + pow(V[i][1] - V[j][1], 2));
			forbidden_arc[i_arc] = false;
			A_square_distances[i][j] = A_lengths[i_arc];

			for (double rho = 0.1; rho < 1.0; rho += 0.1){ // Checking if the arc should be forbidden for falling outside the zone
				arc_middle[0] = rho * V[i][0] + (1 - rho) * V[j][0];
				arc_middle[1] = rho * V[i][1] + (1 - rho) * V[j][1];

				if (((arc_middle[0] > 2) && (arc_middle[0] < 4) && (arc_middle[1] < 8)) || ((arc_middle[0] > 6) && (arc_middle[0] < 8) && (arc_middle[1] > 2))) {
					forbidden_arc[i_arc] = true;
				}
			}
			i_arc++;
		}
	}

	// Sorting the arc list, just bubble sort
	// Forbiddenness first, nonforbidden will be first
	for (int i = 0; i < number_of_nodes * number_of_nodes - 1; i++) {
		for (int j = number_of_nodes * number_of_nodes - 2; j > i - 1; j--) {
			if (forbidden_arc[j] < forbidden_arc[j + 1]) {
				int buffer_0 = A_ij[j][0];
				int buffer_1 = A_ij[j][1];
				bool buffer_forbidden = forbidden_arc[j]; 
				double buf = A_lengths[j];
				// swapping forbidenness
				// swapping A_ijs

				A_ij[j][0] = A_ij[j + 1][0];
				A_ij[j][1] = A_ij[j + 1][1];
				forbidden_arc[j] = forbidden_arc[j + 1];
				A_lengths[j] = A_lengths[j + 1];

				A_ij[j + 1][0] = buffer_0;
				A_ij[j + 1][1] = buffer_1;
				forbidden_arc[j + 1] = buffer_forbidden;
				A_lengths[j + 1] = buf;
			}
		}
	}

	// Length second, why that might revert the forbidenness order? TODO
	for (int i = 0; i < number_of_nodes * number_of_nodes - 1; i++) {
		for (int j = number_of_nodes * number_of_nodes - 2; j > i - 1; j--){
			if ((A_lengths[j] < A_lengths[j + 1]) && (forbidden_arc[j] == forbidden_arc[j + 1])){ // sorting only within the same forbiddenness sets
				int buffer_0 = A_ij[j][0];
				int buffer_1 = A_ij[j][1];
				bool buffer_forbidden = forbidden_arc[j];
				double buf = A_lengths[j];

				A_ij[j][0] = A_ij[j + 1][0];
				A_ij[j][1] = A_ij[j + 1][1];
				forbidden_arc[j] = forbidden_arc[j + 1];
				A_lengths[j] = A_lengths[j + 1];

				A_ij[j + 1][0] = buffer_0;
				A_ij[j + 1][1] = buffer_1;
				forbidden_arc[j + 1] = buffer_forbidden;
				A_lengths[j + 1] = buf;
			}
		}
	}

	
	for (int i = 0; i < number_of_nodes * number_of_nodes; i++) {
		std::cout << A_ij[i][0] << " " << A_ij[i][1] << " " << A_lengths[i] << " " << forbidden_arc[i] << "\n"; // printing stuff after the sorting
	}

	int ** ij_to_arc; // matrix for tranlating (i, j) into arc number
	ij_to_arc = new int *[number_of_nodes];
	for (int i = 0; i < number_of_nodes; i++){
		ij_to_arc[i] = new int[number_of_nodes];
	}

	for (int i_arc = 0; i_arc < number_of_nodes * number_of_nodes; i_arc ++){
		ij_to_arc[A_ij[i_arc][0]][A_ij[i_arc][1]] = i_arc;
	}

	// ARC REMOVAL
	// Table if a given arc is included in the graph still
	bool* A_included; // A_included to change to bool
	A_included = new bool[(int)(number_of_nodes * number_of_nodes)];
	for (int i_arc = 0; i_arc < number_of_nodes * number_of_nodes; i_arc++){
		A_included[i_arc] = true;
	}

	bool** A_square_included;
	A_square_included = new bool* [number_of_nodes];

	for (int i = 0; i < number_of_nodes; i++){
		A_square_included[i] = new bool[number_of_nodes];

		for (int j = 0; j < number_of_nodes; j++){
			A_square_included[i][j] = true;
		}
	}

	int arcs_left = number_of_nodes * number_of_nodes;
	bool removed_arc;

	int * neighbors_in; // what is this
	neighbors_in = new int [number_of_nodes];
	int i_arc_reverse;
	
	// We first remove the arcs starting for the longest first
	for (int i_arc = 0; i_arc < number_of_nodes * number_of_nodes; i_arc += 2) {
		if (forbidden_arc[i_arc]) {
			removed_arc = true; // assume we remove a given arc
			A_included[i_arc] = false; // assume we remove a given arc
			A_included[i_arc + 1] = false; // assume we remove a given arc
			A_square_included[A_ij[i_arc][0]][A_ij[i_arc][1]] = false; // assume we remove a given arc
			A_square_included[A_ij[i_arc + 1][0]][A_ij[i_arc + 1][1]] = false; // assume we remove a given arc

			//std::cout << A_ij[i_arc][0] << " " << A_ij[i_arc][1] << "\n\r";
			//system("pause");

			if (NotConnected(number_of_nodes, A_square_included, A_ij[i_arc][0], A_ij[i_arc][1])) {// If not possible to go from one end to another anymore
				A_included[i_arc] = true; // bring the arc back
				A_included[i_arc + 1] = true; // bring the arc back
				A_square_included[A_ij[i_arc][0]][A_ij[i_arc][1]] = true; // bring the arc back
				A_square_included[A_ij[i_arc + 1][0]][A_ij[i_arc + 1][1]] = true; // bring the arc back
				removed_arc = false;
			}

			if (removed_arc) {
				arcs_left -= 2;
			}

			if (arcs_left < 6 * number_of_nodes) {
				break;
			}
		}
	}

	// New removal of arcs
	for (int i = 0; i < number_of_nodes; i++){ // count the number of neighbors still included
		int how_many_neighbors_in = 0;

		for (int j = 0; j < number_of_nodes; j++){
			if (A_square_included[i][j]){
				neighbors_in[how_many_neighbors_in] = j; // matrix with the neighbors still included
				how_many_neighbors_in++;
			}
		}

		// Sort the neighbors now first sort: forbidden to the end
		for (int neighbor_1 = how_many_neighbors_in - 1; neighbor_1 > 0; neighbor_1--) {
			for (int neighbor_2 = 0; neighbor_2 < neighbor_1; neighbor_2++) {
				int arc_1 = ij_to_arc[i][neighbors_in[neighbor_2]];
				int arc_2 = ij_to_arc[i][neighbors_in[neighbor_2 + 1]];

				// First sort: length
				if ((forbidden_arc[arc_1] && !forbidden_arc[arc_2])) {
					int buffer_0 = neighbors_in[neighbor_2];
					neighbors_in[neighbor_2] = neighbors_in[neighbor_2 + 1];
					neighbors_in[neighbor_2 + 1] = buffer_0;
				}
			}
		}

		// Second sort: length, shortest will be first
		for (int neighbor_1 = how_many_neighbors_in - 1; neighbor_1 > 0; neighbor_1--) {
			for (int neighbor_2 = 0; neighbor_2 < neighbor_1; neighbor_2++) {
				int arc_1 = ij_to_arc[i][neighbors_in[neighbor_2]];
				int arc_2 = ij_to_arc[i][neighbors_in[neighbor_2 + 1]];

				if ((forbidden_arc[arc_1] == forbidden_arc[arc_2]) && (A_lengths[arc_1] > A_lengths[arc_2])) {
					int buffer_0 = neighbors_in[neighbor_2];
					neighbors_in[neighbor_2] = neighbors_in[neighbor_2 + 1];
					neighbors_in[neighbor_2 + 1] = buffer_0;
				}
			}
		}

		int neighbors_left = how_many_neighbors_in; // To see how many neighbors are still there

		for (int neighbor_1 = how_many_neighbors_in - 1; neighbor_1 > - 1; neighbor_1--){
			if (!(i == neighbors_in[neighbor_1])) { // if the neighbor is not the node itself
				i_arc = ij_to_arc[i][neighbors_in[neighbor_1]]; // identifying which arc is it
				i_arc_reverse = ij_to_arc[neighbors_in[neighbor_1]][i];
				removed_arc = true; // assume we remove a given arc
				A_included[i_arc] = false; // assume we remove a given arc
				A_included[i_arc_reverse] = false; // assume we remove a given arc
				A_square_included[A_ij[i_arc][0]][A_ij[i_arc][1]] = false; // assume we remove a given arc
				A_square_included[A_ij[i_arc_reverse][0]][A_ij[i_arc_reverse][1]] = false; // assume we remove a given arc

				if ((A_ij[i_arc][0] == 0) && (A_ij[i_arc][1] == number_of_nodes - 1)) {
					std::cout << arcs_left << " " << NotConnected(number_of_nodes, A_square_included, A_ij[i_arc][0], A_ij[i_arc][1]) << "\n";
				}

				if (NotConnected(number_of_nodes, A_square_included, A_ij[i_arc][0], A_ij[i_arc][1])) { // If not possible to go from one end to another anymore
					A_included[i_arc] = true; // bring the arc back
					A_included[i_arc_reverse] = true; // bring the arc back
					A_square_included[A_ij[i_arc][0]][A_ij[i_arc][1]] = true; // bring the arc back
					A_square_included[A_ij[i_arc_reverse][0]][A_ij[i_arc_reverse][1]] = true; // bring the arc back
					removed_arc = false;
				}

				if (removed_arc) {
					arcs_left -= 2;
					neighbors_left--;
				}

				if (neighbors_left < neighbors_left_threshold) {
					break;
				}
			}
		}
	}

	int** A;
	A = new int* [arcs_left];

	for (int i_arc = 0; i_arc < arcs_left; i_arc++){
		A[i_arc] = new int[2];
	}

	int i_arc_real = 0;
	for (int i = 0; i < number_of_nodes * number_of_nodes; i++){
		if (A_included[i]){
			//std::cout << (i_arc_real + 1) << " out of " << arcs_left << "\n";
			A[i_arc_real][0] = A_ij[i][0];
			A[i_arc_real][1] = A_ij[i][1];
			i_arc_real++;
		}
	}

	// For each arc build a list of nodes that you can go to from it
	int* how_many_neighbors;
	how_many_neighbors = new int[number_of_nodes];
	int** neighbors;
	neighbors = new int* [number_of_nodes];

	delete[] ij_to_arc; // Translate i j to only the list of arcs that are still left
	ij_to_arc = new int* [number_of_nodes];

	for (int j = 0; j < number_of_nodes; j++){
		ij_to_arc[j] = new int[number_of_nodes];
	}

	for (int i = 0; i < number_of_nodes; i++){
		how_many_neighbors[i] = 0;

		for (int j = 0; j < number_of_nodes; j++){
			if (A_square_included[i][j])
			{
				how_many_neighbors[i]++;
			}
		}

		neighbors[i] = new int[how_many_neighbors[i]];

		int iterate_which = 0;
		neighbors[i][iterate_which] = i;
		iterate_which++;

		for (int j = 0; j < number_of_nodes; j++){
			if ((A_square_included[i][j])){
				if (!(j == i)){
					neighbors[i][iterate_which] = j;
					iterate_which++;
				}
			}
		}
	}

	// Constructing the square matrix that stores the numbers of arcs in our list of arcs, corresponding to their posiiton in the matrix
	for (int i_arc = 0; i_arc < arcs_left; i_arc++){
		ij_to_arc[A[i_arc][0]][A[i_arc][1]] = i_arc;
	}

	std::chrono::steady_clock::time_point begin_t; // = std::chrono::steady_clock::now();
	std::chrono::steady_clock::time_point end_t;
	begin_t = std::chrono::steady_clock::now();

	//////////////////////////// FIRST THE STATIC SOLUTIONS //////////////////////////////////////
	GRBEnv* env = 0; // Gurobi environments
	env = new GRBEnv();

	double length_path; // Array to store path lengths
	double static_paths_lengths;
	double crystal_ball_lengths;
    double* solution_times;
	solution_times = new double [2];
	
    if (i_method == 0){
        for (int i_fake = 0; i_fake < 1; i_fake++){
            begin_t = std::chrono::steady_clock::now();
            std::cout << "Gamma: " << Gamma << "\n";
            GRBModel model = GRBModel(*env);
            //model.set(GRB_IntParam_LogToConsole, 0);
            model.set(GRB_StringAttr_ModelName, "RoutingStatic");
            model.set(GRB_DoubleParam_TimeLimit, 7200);
            model.set(GRB_IntAttr_ModelSense, GRB_MINIMIZE);
            model.set(GRB_IntParam_OutputFlag, output_flag);
            model.set(GRB_IntParam_Threads, 4);

            GRBVar* x; // binary variables denoting the path
            x = model.addVars(arcs_left, GRB_BINARY);

            GRBVar* lambda; // dual multipliers of the <= 1 constraints on perturbations
            lambda = model.addVars(arcs_left);

            GRBVar* mu; // dual multipliers of the sum(xi_ij) <= Gamma constraints on perturbations
            mu = model.addVars(1);
            model.addConstr(mu[0] >= 0);
            GRBLinExpr objvar = 0; // Linear expression to store the objective value

            for (int i_arc = 0; i_arc < arcs_left; i_arc++){
                objvar += A_square_distances[A[i_arc][0]][A[i_arc][1]] * x[i_arc];
                objvar += lambda[i_arc];

                model.addConstr(lambda[i_arc] >= 0.0); // Nonnegativity of the multiplier

                if (A[i_arc][0] == A[i_arc][1]){
                    model.addConstr(x[i_arc] == 0); // Do not use (i , i) arcs
                }

                model.addConstr(A_square_distances[A[i_arc][0]][A[i_arc][1]] * x[i_arc] <= mu[0] + lambda[i_arc]); // Constraints of the RC
            }

            objvar += Gamma * mu[0];
            model.setObjective(objvar); // Set the objective

            // Now the balance constraints
            for (int i = 0; i < number_of_nodes; i++){
                if ((i > 0) && (i < number_of_nodes - 1)){// Flow constraints for intermediate nodes
                    GRBLinExpr balance = 0;

                    for (int j = 1; j < how_many_neighbors[i]; j++){
                        balance += (x[ij_to_arc[i][neighbors[i][j]]] - x[ij_to_arc[neighbors[i][j]][i]]);
                    }

                    model.addConstr(balance == 0);

                } else if (i == 0) { // Flow constraints for start node
                    GRBLinExpr balance = 0;

                    for (int j = 1; j < how_many_neighbors[i]; j++){
                        balance += (x[ij_to_arc[i][neighbors[i][j]]] - x[ij_to_arc[neighbors[i][j]][i]]);
                    }

                    model.addConstr(balance == 1);
                } else { // Flow constraints for end node
                    GRBLinExpr balance = 0;

                    for (int j = 1; j < how_many_neighbors[i]; j++){
                        balance += (x[ij_to_arc[i][neighbors[i][j]]] - x[ij_to_arc[neighbors[i][j]][i]]);
                    }

                    model.addConstr(balance == -1);
                }
            }

            // A constraint that out of two opposing arcs, only one can be used
            for (int i = 0; i < number_of_nodes - 1; i++){
                for (int j = i + 1; j < number_of_nodes; j++){
                    if (A_square_included[i][j]){
                        model.addConstr(x[ij_to_arc[i][j]]  + x[ij_to_arc[j][i]] <= 1);
                    }
                }
            }

            model.optimize();
            end_t = std::chrono::steady_clock::now();
            solution_times[0] = (double)(std::chrono::duration_cast<std::chrono::seconds>(end_t - begin_t).count());

            static_paths_lengths = model.get(GRB_DoubleAttr_ObjVal);
            //outfile << model.get(GRB_DoubleAttr_ObjVal) << " ";

            length_path = 0;

            for (int j = 0; j < arcs_left; j++){
                length_path += x[j].get(GRB_DoubleAttr_X); // Compute the path length
            }

            ofstream outfile_static;
            ostringstream o_filename_static;
            o_filename_static << "static_" << number_of_nodes << "_" << i_instance << "_" << Gamma << ".txt";
            outfile_static.open(o_filename_static.str().c_str());

            for (int i_arc = 0; i_arc < arcs_left; i_arc++){
                if (x[i_arc].get(GRB_DoubleAttr_X) > 0.0001){
                    //std::cout << "x[" << A[i_arc][0] << "," << A[i_arc][1] << "] = " << x[i_arc].get(GRB_DoubleAttr_X) << "\n\r";
                    outfile_static << A[i_arc][0] << " " << A[i_arc][1] << "\n";
                }
            }

            outfile_static.close();
        }

        ///////////////////////////////////////////////////////////////////
        // The crystall ball solutions now

		begin_t = std::chrono::steady_clock::now();
        for (int i_fake = 0; i_fake < 1; i_fake++){
            GRBModel model = GRBModel(*env);
            
            model.set(GRB_StringAttr_ModelName, "CrystalBall");
            model.set(GRB_IntAttr_ModelSense, GRB_MAXIMIZE);
            model.set(GRB_IntParam_OutputFlag, output_flag);
            model.set(GRB_IntParam_Threads, 4);

            GRBVar* xi; // perturbations
            xi = model.addVars(arcs_left);

            GRBVar* y; // dual variables for the nodes
            y = model.addVars(number_of_nodes);

            GRBVar* B_per_i; // dual variables for the nodes
            B_per_i = model.addVars(number_of_nodes, GRB_INTEGER);

            GRBLinExpr objvar = 0;
            objvar = 0; // Linear expression to store the objective value
            objvar += -y[0] + y[number_of_nodes - 1]; /// objective function
            model.setObjective(objvar); // Set the objective

            GRBLinExpr total_pertubation = 0;
            GRBLinExpr total_B_per_i = 0;

            for (int i_arc = 0; i_arc < arcs_left; i_arc++){
                model.addConstr(xi[i_arc] >= 0.0);
                model.addConstr(xi[i_arc] <= 1.0);
                total_pertubation += xi[i_arc];

                model.addConstr((1 + xi[i_arc]) * A_square_distances[A[i_arc][0]][A[i_arc][1]] + y[A[i_arc][0]] - y[A[i_arc][1]] >= 0.0);
            }

            model.addConstr(total_pertubation <= Gamma);

            for (int i_node = 0; i_node < number_of_nodes; i_node++){
                total_pertubation = 0;

                for (int j = 0; j < how_many_neighbors[i_node]; j++){
                    total_pertubation += xi[ij_to_arc[i_node][neighbors[i_node][j]]];
                }

                model.addConstr(total_pertubation <= B_per_i[i_node]);
                model.addConstr(0 <= B_per_i[i_node]);
                total_B_per_i += B_per_i[i_node];
            }

            model.addConstr(total_B_per_i <= Gamma); // total perturbation constraint
            model.optimize();
            end_t = std::chrono::steady_clock::now();
            solution_times[1] = (double)(std::chrono::duration_cast<std::chrono::seconds>(end_t - begin_t).count());
            crystal_ball_lengths = model.get(GRB_DoubleAttr_ObjVal);
        }

        // Outputting to the file
        ofstream outfile;
        ostringstream o_filename;
        o_filename << "routing_output_" << number_of_nodes << "_" << i_instance << "_" << i_method << "_" << Gamma << ".txt";

        outfile.open(o_filename.str().c_str());
        outfile << static_paths_lengths << " ";
        outfile << length_path << " ";
        outfile << crystal_ball_lengths << " ";
        outfile << solution_times[0] << " ";
        outfile << solution_times[1] << " ";
        outfile << "\n";
        outfile.close();

	} else { // If method is not static or crystal ball

        //////////////// ADAPTIVE OPTIMIZATION STARTS
        // Array Q to store the cost-to-go
        double*** Q;
        int*** arcs_to_go;
        Q = new double** [(int)(T + 1)];
        arcs_to_go = new int** [(int)(T + 1)];

        for (int i = 0; i < T + 1; i++) {
            Q[i] = new double* [number_of_nodes];
            arcs_to_go[i] = new int* [number_of_nodes];

            for (int j = 0; j < number_of_nodes; j++) {
                Q[i][j] = new double[(int)(max_Gamma + 1)];
                arcs_to_go[i][j] = new int[(int)(max_Gamma + 1)];
            }
        }

        double*** next_arc_cost;
        next_arc_cost = new double** [(int)(T + 1)];
        for (int i = 0; i < T + 1; i++)
        {
            next_arc_cost[i] = new double* [number_of_nodes];

            for (int j = 0; j < number_of_nodes; j++)
            {
                next_arc_cost[i][j] = new double[(int)(max_Gamma + 1)];
            }
        }

        // Array chosen_Q to store the worst-case optimal actions next node chosen
        int*** chosen_Q;
        chosen_Q = new int** [(int)(T + 1)];
        for (int i = 0; i < T + 1; i++){
            chosen_Q[i] = new int* [number_of_nodes];

            for (int j = 0; j < number_of_nodes; j++){
                chosen_Q[i][j] = new int[(int)(max_Gamma + 1)];

                for (int k = 0; k < max_Gamma + 1; k++)
                {
                    chosen_Q[i][j][k] = -1;
                }
            }
        }

        // Array chosen_Q to store if at a state the worst-case is to perturb or not
        int*** perturbed_Q;
        perturbed_Q = new int** [(int)(T + 1)];

        for (int i = 0; i < T + 1; i++){
            perturbed_Q[i] = new int* [number_of_nodes];

            for (int j = 0; j < number_of_nodes; j++){
                perturbed_Q[i][j] = new int[(int)(max_Gamma + 1)];
            }
        }

        //////////////// ADAPTIVE OPTIMIZATION VERSION
        double adaptive_solution_time;
        begin_t = std::chrono::steady_clock::now();

        // Here starts the main loop 
        int max_neighbors = 0;

        for (int i = 0; i < number_of_nodes; i++){
            if (how_many_neighbors[i] > max_neighbors){
                max_neighbors = how_many_neighbors[i];
            }
        }

        for (int t = T; t >= 0; t--) { // Moving backwards in time
            std::cout << "T = " << t << "\n";
            #pragma omp parallel for
            for (int i = 0; i < number_of_nodes; i++) { // For all nodes
                //std::cout << "i = " << i << "\n";
                for (int B = 0; B < max_Gamma + 1; B++) { // For all budgets
                    //std::cout << "T = " << t  << ", i = " << i << ", B = " << B << "\n";
                    if (t == T) // If end of time horizon then Q = 0 if the right node, 10000 otherwise
                    {
                        if (i == number_of_nodes - 1)
                        {
                            Q[t][i][B] = 0;
                            arcs_to_go[t][i][B] = 0;
                            perturbed_Q[t][i][B] = 0;
                        }
                        else
                        {
                            Q[t][i][B] = 10000;
                            arcs_to_go[t][i][B] = 10000;
                            perturbed_Q[t][i][B] = 0;
                        }
                    } else { // If not at the end of the time horizon
                        if (i == number_of_nodes - 1){ // If at the final node, then Q = 0
                            Q[t][i][B] = 0;
                            arcs_to_go[t][i][B] = 0;
                            perturbed_Q[t][i][B] = 0;
                        } else { // If not at the end of the time horizon and NOT  in the final node
                            if (B > max_neighbors * (T - t)){ // Cheat trick: if remainining budget is strictly more than number of max arcs to go still, then just copy the Q for the state for B - 1 budget (it can't go worse)
                                Q[t][i][B] = Q[t][i][B - 1];
                                arcs_to_go[t][i][B] = arcs_to_go[t][i][B - 1];
                                perturbed_Q[t][i][B] = perturbed_Q[t][i][B - 1];
                                chosen_Q[t][i][B] = chosen_Q[t][i][B - 1];
                            } else {// If cheat trick doesn't apply
                                // Case 1: no perturbation
                                double best_no_perturbation_cost = 10000; // Best cost if there's no perturbation
                                bool all_infty_0 = true; // logical if in no-perturbation case all next arcs give cost-to-go infinity
                                arcs_to_go[t][i][B] = 10000;

                                for (int j = 1; j < how_many_neighbors[i]; j++){ //arcs_to_go[t][i][B] = arcs_to_go[t + 1][neighbors[i][j]][B] + 1;
                                    if (A_square_distances[i][neighbors[i][j]] + Q[t + 1][neighbors[i][j]][B] < best_no_perturbation_cost){
                                        best_no_perturbation_cost = A_square_distances[i][neighbors[i][j]] + Q[t + 1][neighbors[i][j]][B];
                                        arcs_to_go[t][i][B] = arcs_to_go[t + 1][neighbors[i][j]][B] + 1;
                                        chosen_Q[t][i][B] = neighbors[i][j];
                                        perturbed_Q[t][i][B] = 0;

                                        all_infty_0 = false; // at least one next arc gives a finite cost to go
                                    }
                                }

                                if (all_infty_0){ // If even in the no perturbation case all neighbors give an infinite cost-to-go
                                    arcs_to_go[t][i][B] = 10000;
                                    chosen_Q[t][i][B] = -1;
                                    perturbed_Q[t][i][B] = 0;
                                }

                                perturbed_Q[t][i][B] = 0;

                                double** perturbation_costs;
                                perturbation_costs = new double* [how_many_neighbors[i]];

                                for (int j = 0; j < how_many_neighbors[i]; j++){
                                    if (B > 1){
                                        perturbation_costs[j] = new double[B];
                                    } else {
                                        perturbation_costs[j] = new double[1];
                                    }
                                }

                                // Case 2: there is a perturbation and not all next arcs give inifinity cost-to-go (otherwise it makes no sense even to check the robust case)
                                if ((B > 0) && !(all_infty_0)){
                                    for (int j = 1; j < how_many_neighbors[i]; j++){ // Check over all possible next arcs (staying not possible)
                                        for (int B_used = 1; B_used < B + 1; B_used++){
                                            if (B_used < how_many_neighbors[i]){ // Tu jest nowe
                                                if ((Q[t + 1][neighbors[i][j]][B - B_used] < 10000)){ // Only solve an optimization problem if next arc gives a chance for a finite cost-to-go
                                                    GRBModel model = GRBModel(*env);
                                                    model.set(GRB_StringAttr_ModelName, "Routing");
                                                    model.set(GRB_IntAttr_ModelSense, GRB_MAXIMIZE);
                                                    model.set(GRB_IntParam_OutputFlag, output_flag);
                                                    model.set(GRB_IntParam_Threads, 1);

                                                    GRBVar* xi = 0;
                                                    xi = model.addVars(how_many_neighbors[i] - 1); // Perturbation vector variable
                                                    model.set(GRB_IntParam_LogToConsole, 0);

                                                    GRBLinExpr objvar = 0;
                                                    objvar += A_square_distances[i][neighbors[i][j]];
                                                    objvar += A_square_distances[i][neighbors[i][j]] * xi[j - 1];
                                                    objvar += Q[t + 1][neighbors[i][j]][B - B_used];
                                                    model.setObjective(objvar);
                                                    GRBLinExpr ntot = 0;

                                                    for (int l = 0; l < how_many_neighbors[i] - 1; l++){
                                                        model.addConstr(xi[l] >= 0.0); // [0, 1] constraints on perturbations
                                                        model.addConstr(xi[l] <= 1.0); // ....
                                                        ntot += xi[l]; // summing all pertubations
                                                    }

                                                    model.addConstr(ntot <= B_used); // sum of perturbations <= B_used

                                                    // Constraint all other next arcs are at most as good

                                                    for (int k = 1; k < how_many_neighbors[i]; k++)
                                                    {
                                                        if (!(k == j))
                                                        {
                                                            GRBLinExpr LeftHand = 0; // Linear expressions to store the Left- and Right-hand-sides in the constraint of the recursion problem
                                                            GRBLinExpr RightHand = 0;

                                                            LeftHand += Q[t + 1][neighbors[i][j]][B - B_used];
                                                            RightHand += Q[t + 1][neighbors[i][k]][B - B_used];

                                                            LeftHand += A_square_distances[i][neighbors[i][j]];
                                                            RightHand += A_square_distances[i][neighbors[i][k]];

                                                            LeftHand += xi[j - 1] * A_square_distances[i][neighbors[i][j]];
                                                            RightHand += xi[k - 1] * A_square_distances[i][neighbors[i][k]];

                                                            model.addConstr(LeftHand <= RightHand);
                                                        }
                                                    }

                                                    model.optimize();

                                                    if (model.get(GRB_IntAttr_Status) == GRB_OPTIMAL) // If problem was feasible, i.e., node j is best for at least some uncertainty outcomes
                                                    {
                                                        perturbation_costs[j][B_used - 1] = model.get(GRB_DoubleAttr_ObjVal);
                                                    }
                                                    else // If node j gives a finite cost but is not best for any uncertainty outcome then just don't take it into account
                                                    {
                                                        perturbation_costs[j][B_used - 1] = -1;
                                                    }

                                                } else {
                                                    perturbation_costs[j][B_used - 1] = 10000; // If every next node gives infinity cost-to-go then set cost to go as infinity
                                                }
                                            } else {
                                                perturbation_costs[j][B_used - 1] = perturbation_costs[j][how_many_neighbors[i] - 2];
                                            }
                                        }
                                    }
                                }

                                Q[t][i][B] = best_no_perturbation_cost; // Tentatively set cost-to-go as the no-perturbation one

                                if ((B > 0) && !(all_infty_0)){ // If there is uncertainty budget left and cost-to-go finite in no-uncertainty (easier) situation, check the uncertainty-driven costs
                                    bool all_infty = true; // Boolean to check if in the perturbed case all next arcs can lead to infinity costs
                                    for (int j = 1; j < how_many_neighbors[i]; j++){
                                        if (perturbation_costs[j][1] < 10000){
                                            all_infty = false;
                                            break;
                                        }
                                    }

                                    if (all_infty){ // if in the perturbed case all next arcs can lead to infinity costs then set the cost as infinity
                                        Q[t][i][B] = 10000;
                                        arcs_to_go[t][i][B] = 10000;
                                        chosen_Q[t][i][B] = -1;
                                        perturbed_Q[t][i][B] = 1;
                                    } else {// if in the perturbed case at least one next arc  gives a finite worst-case cost then check them
                                        for (int j = 1; j < how_many_neighbors[i]; j++){
                                            for (int B_used = 1; B_used < B + 1; B_used++){
                                                if ((perturbation_costs[j][B_used - 1] > Q[t][i][B]) && (perturbation_costs[j][B_used - 1] < 10000)) // if in the perturbed case arc j gives a finite worst-case cost then, and this is worse than in the no-perturbation case
                                                {
                                                    Q[t][i][B] = perturbation_costs[j][B_used - 1];
                                                    arcs_to_go[t][i][B] = arcs_to_go[t + 1][neighbors[i][j]][B - B_used] + 1;
                                                    perturbed_Q[t][i][B] = 1;
                                                    chosen_Q[t][i][B] = neighbors[i][j];
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                    //std::cout << "Q[" << t << "][" << i << "][" << B << "]=" << Q[t][i][B] << "\n";
                }
            }
        }

        end_t = std::chrono::steady_clock::now();
        adaptive_solution_time = (double)(std::chrono::duration_cast<std::chrono::seconds>(end_t - begin_t).count());

        ofstream outfile;
        ostringstream o_filename;
        o_filename << "routing_output_" << number_of_nodes << "_" << i_instance << "_" << i_method << ".txt";

        outfile.open(o_filename.str().c_str());
        for (int i_Gamma = 0; i_Gamma < max_Gamma + 1; i_Gamma++){
            outfile << Q[0][0][i_Gamma] << " ";
            outfile << arcs_to_go[0][0][i_Gamma] << " ";
            outfile << adaptive_solution_time;
            outfile << "\n";
        }
        outfile.close();

        // Simulation of what can happen, do pertrubations only on the "real arcs"
        int* arc_usage_count;// Number of arcs to run the statistics on
        arc_usage_count = new int[arcs_left];
        int* budget_per_node;
        budget_per_node = new int[number_of_nodes];
        double* perturbations_per_arc;
        perturbations_per_arc = new double[number_of_nodes];
        ostringstream o_filename_arcs;
        ofstream outfile_arcs;

        //double arc_probability = 0.1;
        double portion_size = 0.1;
        int arc_index = 0;
        bool found_the_portion_a_place = false;
        int implemented_budget_to_perturb;
        int budget_to_perturb;
        double total_budget_left = 0;
        int current_node = 0;
        int current_B;
        int real_current_B;
        int current_T = 0;
        double sum_perturbations = 0;
        double rv;
        double* real_costs;
        real_costs = new double[number_of_nodes];

        for (int i_B = 0; i_B <= max_Gamma; i_B++) {
            //std::cout << "arc_probability: " << arc_probability << "\n";
            double stored_p;

            for (double arc_probability = 0.2; arc_probability > 0; arc_probability -= 0.005) {
                stored_p = arc_probability;
                //std::cout << "i_B: " << i_B << "\n";
                //std::cout << "Simulation for i_B=" << i_B << "\n";
                total_budget_left = 0.0;

                // Proper simulation
                for (int i_arc = 0; i_arc < arcs_left; i_arc++) {
                    arc_usage_count[i_arc] = 0; // zero the arc usage count
                }

                //std::cout << "stage1\n";

                for (int i_sim = 0; i_sim < N_sim; i_sim++) { // Iterate over simulations
                    //std::cout << "i_sim= " << i_sim << "\n";
                    current_node = 0;
                    current_B = i_B;
                    real_current_B = i_B;
                    current_T = 0;
                    //std::cout << "time: " << current_T << " node: " << current_node << " B= " << current_B << "\n";

                    while (!(current_node == number_of_nodes - 1)) { // && (i_count < 30)
                        // Computing the perturbations
                        // 1) Just sampling the direction
                        //std::cout << "Nb neighbors: " << how_many_neighbors[current_node] << "\n";
                        budget_to_perturb = 0;

                        // How much budget will be perturbed
                        budget_to_perturb = 0;
                        for (int i_neighbor = 1; i_neighbor < how_many_neighbors[current_node]; i_neighbor++) {
                            rv = ((double)rand() / (RAND_MAX));
                            if (rv < arc_probability) {
                                budget_to_perturb++;
                            }
                        }
                        //std::cout << "stage2\n";

                        implemented_budget_to_perturb = fmin(budget_to_perturb, real_current_B);
                        for (int i_arc = 1; i_arc < how_many_neighbors[current_node]; i_arc++) {
                            perturbations_per_arc[i_arc] = 0.0;
                        }
                        //std::cout << "stage3\n";

                        // Now allocating the perturbation
                        for (int i_portion = 0; i_portion < (implemented_budget_to_perturb / portion_size); i_portion++) {
                            found_the_portion_a_place = false;
                            while (!found_the_portion_a_place) {
                                arc_index = 1 + (int)(floor(((double)rand() / (RAND_MAX)) * ((double)(how_many_neighbors[current_node]) - 1)));
                                if(arc_index == how_many_neighbors[current_node]){
                                    std::cout << "What should not happen, happened\n";
                                    arc_index == how_many_neighbors[current_node] - 1;
                                }

                                if (perturbations_per_arc[arc_index] < 1.0 - 0.0001) {
                                    perturbations_per_arc[arc_index] += portion_size;
                                    found_the_portion_a_place = true;
                                }
                            }
                        }
                        //std::cout << "stage4\n";

                        // Computing the total costs of each node
                        //std::cout << "nb neighbors: " << how_many_neighbors[current_node] << "\n";
                        for (int i_neighbor = 1; i_neighbor < how_many_neighbors[current_node]; i_neighbor++) {
                            real_costs[i_neighbor] = A_square_distances[current_node][neighbors[current_node][i_neighbor]] * (1 + perturbations_per_arc[i_neighbor]) + Q[current_T][neighbors[current_node][i_neighbor]][real_current_B - implemented_budget_to_perturb];
                        }
                        //std::cout << "stage5\n";

                        // Choose the next node
                        double min_distance = 10000;
                        int best_neighbor = -1;

                        for (int i_neighbor = 1; i_neighbor < how_many_neighbors[current_node]; i_neighbor++) {
                            if (real_costs[i_neighbor] < min_distance) {
                                best_neighbor = neighbors[current_node][i_neighbor];
                                min_distance = real_costs[i_neighbor];
                            }
                        }
                        //std::cout << "stage6\n";

                        budget_per_node[current_node] = 0;
                        arc_usage_count[ij_to_arc[current_node][best_neighbor]]++;
                        current_B -= budget_to_perturb;
                        real_current_B -= implemented_budget_to_perturb;
                        current_node = best_neighbor;
                        current_T += 1;
                        //std::cout << "stage7\n";
                        //std::cout << "time: " << current_T << " node: " << current_node << " B= " << current_B << "\n";
                    }

                    //std::cout << "budget left: " << current_B << "\n";
                    total_budget_left += (double)(current_B);
                }

                if (total_budget_left / (double)(N_sim) > 0){
                    break;
                }
            }

            o_filename_arcs.str("");
            o_filename_arcs.clear();
            o_filename_arcs << "arc_statistics_" << number_of_nodes << "_" << i_instance << "_" << i_B << ".txt";
            outfile_arcs.open(o_filename_arcs.str().c_str());
            outfile_arcs << total_budget_left / (double)(N_sim) << " " << stored_p << "\n";

            for (int i_arc = 0; i_arc < arcs_left; i_arc++) {
                //std::cout << "[" << A[i_arc][0] << "][" << A[i_arc][1] << "]=" << arc_usage_count[i_arc] << "\n";
                outfile_arcs << A[i_arc][0] << " " << A[i_arc][1] << " " << arc_usage_count[i_arc] << "\n";
            }

            outfile_arcs.close();
        }
    }

	return 0;
}
